from typing import Union, List, Callable

from .....utils import extract_segments, filter_kwargs, raise_invalid_params
from .....io import EventBlock
from ...prompt_template import PromptTemplate
from ...selector import Selector, End
from ..base import BaseAgent
from ..chat import ChatAgent
from .base import FlowAgent

import json

class PlanAndSolve(FlowAgent):
    """
    PlanAndSolve 类似于 ReAct, 一边制定总体计划一边执行。
    这既可以通过一步步思考获得稳定推理, 又可以在每次制定更有利于实现总体目标的计划。
    """
    @classmethod
    def allowed_params(cls):
        return {
            "planner": "计划者",
            "worker": "执行者",
            "replanner": "重新计划者",
            "planner_template": "计划者模板",
            "worker_template": "执行者模板",
            "replanner_template": "重新计划者模板",
            **FlowAgent.allowed_params(),
        }
        
    def __init__(
        self,
        planner: ChatAgent,
        worker: ChatAgent,
        replanner: ChatAgent,
        planner_template: PromptTemplate=None,
        worker_template: PromptTemplate=None,
        replanner_template: PromptTemplate=None,
        **kwargs
    ):
        raise_invalid_params(kwargs, self.allowed_params())

        if not isinstance(planner, ChatAgent) or not isinstance(worker, ChatAgent) or not isinstance(replanner, ChatAgent):
            raise ValueError("planner, worker, replanner 必须是 ChatAgent 的子类")
        if not planner.tools:
            raise ValueError("planner 必须包含可用的工具")
        if planner is worker or planner is replanner or worker is replanner:
            raise ValueError("planner, worker, replanner 不能相同")

        planner_template = planner_template or PromptTemplate("FLOW/PlanAndSolve/Planner")
        self.planner_template = planner_template

        replanner_template = replanner_template or PromptTemplate("FLOW/PlanAndSolve/RePlanner")
        self.replanner_template = replanner_template

        worker_template = worker_template or PromptTemplate("FLOW/PlanAndSolve/Worker")
        self.worker_template = worker_template

        planner.tools_behavior = "parse"
        planner.reset_init_memory(planner_template)
        self.planner = planner

        worker.tools_behavior = "parse-execute"
        worker.reset_init_memory(worker_template)
        self.worker = worker

        replanner.tools_behavior = "parse"
        replanner.reset_init_memory(replanner_template)
        self.replanner = replanner

        self.completed_work = []
        self.todo = []

        def observe_worker_for_replanner(agent: BaseAgent):
            replanner.memory.clear()
            self.todo.clear()

            if agent.tools_calling_steps:
                for step in agent.tools_calling_steps:
                    item = f'Step{step["index"]}: {step["description"]}\n{step["eid"]} = {step["name"]}[{step["arguments"]}]\n'
                    item = f'{item}{step["eid"]} 执行结果: {step["result"]}\n'
                    self.completed_work.append(item)

            self.replanner_template.bind_provider({
                "plan_todo": "\n".join(self.todo),
                "plan_done": "\n".join(self.completed_work)
            })

            yield EventBlock("final_text", self.task)

        def observe_plan_for_worker(agent: BaseAgent):
            worker.memory.clear()
            if agent.tools_calling_steps:
                for step in agent.tools_calling_steps:
                    item = f'Step{step["index"]}: {step["description"]}\n{step["eid"]} = {step["name"]}[{step["arguments"]}]\n'
                    self.todo.append(item)

            self.worker_template.bind_provider( {
                "plan_todo": "\n".join(self.todo)
            })

            yield EventBlock("final_text", "请开始\n")

        def should_continue(vars, runs):
            if self.replanner.final_answer:
                return "__END__"
            else:
                self.replanner.clear()
                return "observe_plan_for_worker"

        super().__init__(
            {"planner": planner},
            {"observe_plan_for_worker": observe_plan_for_worker},
            {"worker": worker},
            {"observe_worker_for_replanner": observe_worker_for_replanner},
            {"replanner": replanner},
            Selector(condition=should_continue, name="should_continue"),
            **filter_kwargs(kwargs, self.allowed_params())
        )

    def begin_call(self, args):
        self.planner.clear()
        self.replanner.clear()
        self.worker.clear()
        self.completed_work.clear()
        self.todo.clear()

        return super().begin_call(args)

    def end_call(self):
        self._last_output = self.replanner.final_answer



from typing import Union, List

from ...utils import raise_invalid_params, filter_kwargs
from ...io import EventBlock, NewLineBlock
from ...core.runnable.agent import ChatAgent
from ...core.runnable.message import Messages
import time

class FakeLLM(ChatAgent):
    @classmethod
    def allowed_params(cls):
        return {
            "response": "响应内容",
            "sleep": "睡眠时间",
            **ChatAgent.allowed_params(),
        }

    def __init__(self, response: Union[str, List[str]]=None, sleep: float=None, **kwargs):
        raise_invalid_params(kwargs, self.allowed_params())
        super().__init__(threads_group="FAKE_LLM", **filter_kwargs(kwargs, self.allowed_params()))

        self.sleep = sleep if sleep is not None else 0.01
        self.response = response if isinstance(response, list) else ([response] if response else None)
        self.current_response_index = 0

    def generate(self, prompt: Union[str, List[str]], sleep: float=0, **kwargs):
        yield EventBlock("info", f'I am FakeLLM')

        _sleep = self.sleep if sleep <= 0 else sleep

        if self.response:
            resp = self.response[self.current_response_index]
            self.current_response_index = (self.current_response_index + 1) % len(self.response)
        else:
            std_msg = Messages(prompt, style="text")
            resp = f"Reply >> {std_msg.last_content}"

        for content in resp:
            time.sleep(_sleep)
            yield EventBlock("chunk", content)
        yield NewLineBlock()



from .python_code import PythonAgent

class PandasAgent(PythonAgent):
    """
    对所提供的 pandas 数据集做分析和处理。
    """

    def __init__(self, *args, template_id: str=None, **kwargs):
        self.template_id = template_id or "CODE/Pandas"

        # 生成作为工具被使用时的功能描述
        _tool_params = kwargs.pop("tool_params", {
            "question": "细致描述数据分析任务的需求描述",
        })
        super().__init__(*args, tool_params=_tool_params, **kwargs)




from typing import Union, List, Optional
from urllib.parse import urlparse
import os
import time

from ..http import EventBlock, save_resource, confirm_base64_or_uri
from ...types import BaseAgent
from ...utils import raise_invalid_params
from ...config import get_env

CHECK_RESULT_SECONDS = get_env("HTTP_CHECK_RESULT_SECONDS")

class CogVideoX(BaseAgent):
    """    
    [CogVideoX API](https://open.bigmodel.cn/dev/api/videomodel/cogvideox)
    """
    @classmethod
    def allowed_params(cls):
        return {
            "model": "模型名称",
            "api_key": "API_KEY",
            "base_url": "API_BASE_URL",
            **BaseAgent.allowed_params()
        }

    def __init__(self, model: str=None, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())
        try:
            from zhipuai import ZhipuAI
        except ImportError:
            raise ImportError(
                "Could not import zhipuai package. "
                "Please install it via 'pip install -U zhipuai'"
            )

        super().__init__(threads_group="COGVIEW", **kwargs)

        self.default_call_args = {
            "model": model or "cogvideox"
        }
        self.model_args = {
            "api_key": kwargs.get("api_key", os.getenv("ZHIPUAI_API_KEY")),
            "base_url": kwargs.get("base_url", os.getenv("ZHIPUAI_BASE_URL")),
        }
        self.client = ZhipuAI(**self.model_args)

        self.description = "我擅长根据你的文字提示描述生成视频，你必须在 prompt 中详细描述生成要求，你必须展开描述，比如镜头、光线、细部等。"
        self.tool_params = {
            "prompt": "请尽量详细描述生成要求的细节",
            "image_url": "提供基于其生成内容的图像。如果传入此参数，系统将以该图像为基础进行操作。",
            "output": "指定生成视频的名称，应当包括扩展名"
        }

    def call(
        self, 
        prompt: str=None,
        image_url: str=None,
        output: Optional[Union[str, List[str]]] = None,
        **kwargs
    ):
        image_url = confirm_base64_or_uri(image_url)

        _kwargs = self.default_call_args
        _kwargs.update({
            "prompt": prompt,
            "image_url": image_url,
            **kwargs,
        })
        if isinstance(output, str):
            output = [output]

        resp = self.client.videos.generations(**_kwargs)
        while resp.task_status == 'PROCESSING':
            resp = self.client.videos.retrieve_videos_result(id=resp.id)
            yield EventBlock("task_status", f'{resp.id} - {resp.task_status}')

            if resp.task_status == 'SUCCESS':
                for result_index, result in enumerate(resp.video_result):
                    url = result.url
                    cover_url = result.cover_image_url
                    yield EventBlock("video_url", url)
                    if not output or result_index >= len(output):
                        parsed_url = urlparse(url)
                        filename = os.path.basename(parsed_url.path)
                        output_path = f"{filename.rsplit('.', 1)[0]}.{filename.rsplit('.', 1)[1]}"
                    else:
                        output_path = output[result_index]
                    yield from save_resource(url, output_path)
                    yield from save_resource(cover_url, output_path + ".png")

            time.sleep(CHECK_RESULT_SECONDS)



from typing import Any, Dict
import pandas as pd
import textwrap

class Dataset:
    """
    数据集。

    默认仅支持 pandas.DataFrame 类型的数据集。可以通过子类继承支持其他类型的数据集。
    """
    def __init__(self, df: pd.DataFrame, name: str, desc: str=None):
        if not desc:
            desc = ""

        if not isinstance(df, pd.DataFrame):
            raise ValueError(f"df 必须是 pandas.DataFrame 类型")
        if desc and not isinstance(desc, str):
            raise ValueError(f"desc 必须是 str 类型")

        self.name = name
        self.df = df
        self.desc = desc or f"关于`{self.name}`的数据集"

    def __repr__(self):
        return f"Dataset(name={self.name}, df={self.df}, desc={self.desc})"

    @property
    def summary(self):
        from .runnable.prompt_template import PromptTemplate

        head = self.df.head()
        example_md = head.to_markdown(index=False)
        summary = PromptTemplate(
            "CODE/DATASET_SUMMARY",
            binding_map={
                "name": lambda: self.name,
                "desc": lambda: self.desc,
                "example_md": lambda: example_md,
            }).format()
        return summary



import os
import fnmatch
import json
import re
from typing import Union, List
from ..runnable import Runnable
from ..document import Document
from ...io import EventBlock
from ...config import get_env
from ...utils import minify_text, count_tokens, raise_invalid_params, filter_kwargs
import numpy as np

class MarkMeta(Runnable):
    """
    MarkMeta 持久化时为基于 Markdown 语法的纯文本，但增加了一些扩展标签；
    加载到内存时主要属性为 Document 列表，并将标签转化为 Document 元素的 meta 键值数据。

    MarkMeta 文件中的标签语法：
    - 扩展标签都独占一行，且以 @ 开头，行首不允许多余空格
    - 主要标签包括 @meta @index @end
    - @meta 包含键值描述语法：<!-- @meta {"k1":v1, "k2":v2, …} -->
    - MarkMeta 文档被 @meta 标签分割成多个片段，从开头或 @meta 标签的下一行开始，到下一个 @meta 标签前一行结束
    - 其他标签智能包含在 @meta 标签中

    MarkMeta 导入和导出：
    - 处理导入
        - 按 @meta 标签切分文档，每个片段单元作为独立的 Document 元素
        - 如果单个文档内容超过 chunk_size 限制，则滚动切分，并增补 ID 作为新的 Document 元素
        - 优先使用 @meta 所在行的键值描述来初始化 Document 的 meta 数据
        - 片段单元的内容都作为 Document 的 text 属性
    - @TODO: 处理 @index 标签
        - Document 的 index 属性是一个列表，额外存储了专门指定的文本嵌入内容，建立索引时应当替代 text 内容
        - 可以在文档中使用 @index 指定开始位置，遇到@end结束
        - @index 支持多段落
    - 处理导出时
        - 将 Document 的 text 属性作为文本内容，meta 作为元数据
        - 由于 text 属性中仍然包含可能存在的 @index 标签，所以导出时不必专门保存 index 内容

    在向量模型、向量数据库中使用 MarkMeta 文件:
    - 导入时按照 Document 元素作为切分单元

    导出时，由对话模型将输出结果保存为 MarkMeta 文件。
    """
    @classmethod
    def allowed_params(cls):
        return {
            "dir": "从这个目录路径导入文件，默认为 ILLUFLY_DOCS 环境变量",
            "filter": "文件名过滤器，可以直接写文件名，或者使用 * 号等通配符",
            "exts": "文件扩展名列表，默认支持 md, Md, MD, markdown, MARKDOWN 等",
            "chunk_size": "每个块的大小，这可能是各个模型处理中对 token 限制要求的，默认 1024",
            "chunk_overlap": "每个块的覆盖大小，默认 100",
            **Runnable.allowed_params()
        }

    def __init__(self, dir: str=None, filter: str=None, exts: list = None, chunk_size: int=None, chunk_overlap: int=None, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        super().__init__(**kwargs)
        self.directory = dir or get_env("ILLUFLY_DOCS")
        self.filename_filter = filter or '*'
        self.extensions = exts or ['*.md', '*.Md', '*.MD', '*.markdown', '*.MARKDOWN']
        self.chunk_size = chunk_size or 1024
        self.chunk_overlap = chunk_overlap or 100
        self.documents = []

    def clear(self):
        self.documents.clear()

    @property
    def last_output(self):
        return self.documents
    
    def save(self) -> List[Document]:
        """
        将文档保存为 MarkMeta 文件。
        """
        source_files = {}
        for doc in self.documents:
            k = doc.meta['source']
            if source_files.get(k) is None:
                source_files[k] = []
            source_files[k].append(doc)
        for source, docs in source_files.items():
            path = os.path.join(get_env("ILLUFLY_DOCS"), source)
            if not os.path.exists(path):
                os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, 'w', encoding='utf-8') as f:
                for doc in docs:
                    # 将 meta 字典转换为 JSON 字符串
                    # 删除 embeddings 键值对，以避免保存向量数据
                    doc.meta.pop('embeddings', None)
                    meta_line = f"@meta {json.dumps(doc.meta, ensure_ascii=False)}"
                    f.write("\n<!-- " + meta_line + " -->\n")
                    f.write(doc.text + "\n")
                yield EventBlock("info", f"Saved file {source} with {len(docs)} chunks")

    def call(self, *files, **kwargs):
        yield from self.load(*files, **kwargs)

    def load(self, **kwargs):
        """
        load 方法是加载文档时最常用的方法。
        """
        self.documents.clear()

        files = self.get_files(self.directory, self.filename_filter, self.extensions)
        for file_path in files:
            try:
                yield from self.load_file(file_path)
            except Exception as e:
                yield(EventBlock("warn", f"读取文件失败 {file_path}: {e}"))

    def get_files(self, path, filename_filter, extensions):
        """
        获取目录下所有符合条件的文件。
        """
        matches = []
        for root, dirnames, filenames in os.walk(path):
            dirnames[:] = [d for d in dirnames if not d.startswith('.')]  # 排除隐藏文件夹
            filenames = [f for f in filenames if not f.startswith('.')]  # 排除隐藏文件
            for extension in extensions:
                for filename in fnmatch.filter(filenames, filename_filter + extension):
                    matches.append(os.path.join(root, filename))
        return matches

    def load_file(self, abs_path: str) -> List[Document]:
        """
        加载单个文件。
        """
        if not abs_path or not os.path.exists(abs_path):
            yield(EventBlock("warn", f"文件不存在 {abs_path}"))

        with open(abs_path, 'r', encoding='utf-8') as f:
            txt = f.read()
            if str(txt).strip() == "":
                yield(EventBlock("warn", f"文件内容为空 {abs_path}"))
                return
            self.load_text(txt, source=abs_path)
            yield(EventBlock("info", f"已成功加载文件 {abs_path} ，其中包含 {len(self.documents)} 个片段。"))

    def load_text(self, text: str, source: str=None) -> List[Document]:
        """
        将文本内容加载为 MarkMeta 文档。
        """
        docs = self.split_with_meta(text)
        for doc in docs:
            if doc.meta.get('source') is None:
                new_source = source or f"{docs[0].meta.get('id', 'unknown')}.md"
                doc.meta["source"] = new_source

            if count_tokens(doc.text) > self.chunk_size:
                chunks = self.split_text_recursive(doc.text, doc.meta['source'])
                self.documents.extend(chunks)
            else:
                self.documents.append(doc)
        return self.documents

    def split_with_meta(self, text: str) -> List[Document]:
        """
        按照 @meta 标签切分文档，每个片段单元作为独立的 Document 元素，并提取元数据。
        """
        documents = []
        split_text = re.split(r'\s*<!--\s*@meta', "\n" + text)
        for segment in split_text:
            if segment.strip() == "":
                continue
            lines = segment.split("\n")
            meta_line = lines[0].strip().replace("<!--", "").replace("-->", "").strip()
            content = "\n".join(lines[1:]).strip()

            try:
                # 直接将 meta_line 作为 JSON 解析
                meta = json.loads(meta_line)
            except json.JSONDecodeError as e:
                meta = {"raw_meta": meta_line}

            doc = Document(text=content, meta=meta)
            documents.append(doc)
        return documents

    def split_text_recursive(self, text: str, source: str) -> List[Document]:
        """
        按照指定规则分割Markdown文档。

        :return: 分割后Document对象列表
        """
        def split_text(text: str) -> List[str]:
            return text.split('\n')

        def create_chunk(lines: List[str], chunk_index: int) -> Document:
            return Document(text='\n'.join(lines), meta={"source": f"{source}#{chunk_index}"})

        chunks = []

        if not isinstance(text, str):
            raise ValueError("split_markdown的参数 text 必须是字符串")

        if not text or text.strip() == "":
            return chunks

        lines = split_text(text)
        current_chunk = []
        current_length = 0

        chunk_index = 0
        for line in lines:
            line_length = count_tokens(line)
            if line_length > self.chunk_size:
                continue  # 忽略超过chunk_size的最小单位

            if current_length + line_length > self.chunk_size:
                docs = [d for (l, d) in current_chunk]
                chunks.append(create_chunk(docs, chunk_index))
                
                # 计算重叠部分
                overlap_length = 0
                overlap_chunk = []
                for l, d in reversed(current_chunk):
                    overlap_length += l
                    overlap_chunk.insert(0, d)
                    if overlap_length >= self.chunk_overlap or overlap_length >= self.chunk_size / 2:
                        break

                current_chunk = [(count_tokens(d), d) for d in overlap_chunk]
                current_length = sum(l for l, _ in current_chunk)
                continue

            current_chunk.append((line_length, line))
            current_length += line_length
        if current_chunk:
            docs = [d for l, d in current_chunk]
            chunks.append(create_chunk(docs, chunk_index))

        return chunks



import json
import copy
import asyncio
import uuid

from abc import abstractmethod
from typing import Union, List, Tuple, Any

from .....config import get_env
from .....utils import merge_tool_calls, extract_text, raise_invalid_params, filter_kwargs
from .....io import EventBlock, EndBlock, NewLineBlock
from ....document import Document
from ...base import Runnable
from ...message import Messages
from ...prompt_template import PromptTemplate
from ..base import BaseAgent
from ..knowledge_manager import KnowledgeManager
from .tools_calling import BaseToolCalling, OpenAIToolsCalling
from .tools_manager import ToolsManager
from .memory_manager import MemoryManager

class ChatAgent(BaseAgent, KnowledgeManager, MemoryManager, ToolsManager):
    """
    对话智能体是基于大模型实现的智能体，可以用于对话生成、对话理解等场景。

    ChatAgent 类包含一些核心属性，用于保存对话过程中的中间结果，例如：
    - task：用户发起的最初任务，这是 _task 的只读版本
    - final_answer：对话过程中输出的最终答案，这是 _final_answer 的只读版本
    - last_output：对话过程中上次调用输出的结果，这是 _last_output 的只读版本

    由于 final_answer 必须经过提示语引导，按照约定格式输出，因此可将 T/FA(task/final_answer) 用于自动化提取 Q/A 预料。
    而 T/FA 的提取结果可被保存在 __XP__ 等经验目录中，在其他对话时被当作检索资料调用。
    这就实现了基本的自我进化过程，而「自我进化」也是 illufly 设计的一个核心目标。
    """
    @classmethod
    def allowed_params(cls):
        """
        返回当前可用的参数列表。
        """
        return {
            "end_chk": "是否在最后输出一个 EndBlock",
            "start_marker": "开始标记，默认为 ```",
            "end_marker": "结束标记，默认为 ```",
            "fetching_context": "上下文提取标记，可通过修改环境变量 ILLUFLY_CONTEXT_START 和 ILLUFLY_CONTEXT_END 修改默认值",
            "fetching_final_answer": "最终答案提取标记，可通过修改环境变量 ILLUFLY_FINAL_ANSWER_START 和 ILLUFLY_FINAL_ANSWER_END 修改默认值",
            **BaseAgent.allowed_params(),
            **KnowledgeManager.allowed_params(),
            **ToolsManager.allowed_params(),
            **MemoryManager.allowed_params(),
        }

    def __init__(
        self,
        end_chk: bool = False,
        start_marker: str=None,
        end_marker: str=None,
        fetching_context: Tuple[str, str]=None,
        fetching_final_answer: Tuple[str, str]=None,
        **kwargs
    ):
        """
        对话智能体支持的核心能力包括：
        - tools：工具调用
        - memory：记忆管理
        - knowledge：知识管理
        """
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        kwargs["tool_params"] = kwargs.get("tool_params", {"prompt": "详细描述用户问题"})
        BaseAgent.__init__(self, **filter_kwargs(kwargs, BaseAgent.allowed_params()))
        KnowledgeManager.__init__(self, **filter_kwargs(kwargs, KnowledgeManager.allowed_params()))
        ToolsManager.__init__(self, **filter_kwargs(kwargs, ToolsManager.allowed_params()))

        self.end_chk = end_chk

        self.start_marker = start_marker or "```"
        self.end_marker = end_marker or "```"

        self.fetching_context = fetching_context or (get_env("ILLUFLY_CONTEXT_START"), get_env("ILLUFLY_CONTEXT_END"))
        self.fetching_final_answer = fetching_final_answer or (get_env("ILLUFLY_FINAL_ANSWER_START"), get_env("ILLUFLY_FINAL_ANSWER_END"))

        # 在子类中应当将模型参数保存到这个属性中，以便持久化管理
        self.model_args = {"base_url": None, "api_key": None}
        self.default_call_args = {"model": None}

        # 增加的可绑定变量
        self._task = ""
        self._final_answer = ""
        self._tools_to_exec = self.get_tools()
        self._resources = ""

        MemoryManager.__init__(self, **filter_kwargs(kwargs, MemoryManager.allowed_params()))

    def clear(self):
        self.memory.clear()
        self._task = ""
        self._final_answer = ""
        self._last_output = ""

    @property
    def task(self):
        return self._task

    @property
    def final_answer(self):
        return self._final_answer

    @property
    def runnable_info(self):
        info = super().runnable_info
        info.update({
            "thread_id": self.thread_id,
            "model_name": self.default_call_args.get("model"),
            **self.model_args
        })
        return info

    @property
    def tools_calling_steps(self):
        """
        返回所有工具回调的计划和中间步骤
        """
        steps = []
        for h in self.tools_handlers:
            steps.extend(h.steps)
        return steps

    @property
    def provider_dict(self):
        local_dict = {
            "task": self.task,
            "final_answer": self.final_answer,
            "tools_calling_steps": self.tools_calling_steps,
            "tools_name": ",".join([a.name for a in self._tools_to_exec]),
            "tools_desc": "\n".join(json.dumps(t.tool_desc, ensure_ascii=False) for t in self._tools_to_exec),
            "knowledge": self.get_knowledge(self.task),
            "recalled_knowledge": self.recalled_knowledge,
            "chat_memory": self.chat_memory
        }
        return {
            **super().provider_dict,
            **{k:v for k,v in local_dict.items() if v is not None},
        }

    @abstractmethod
    def generate(self, prompt: Union[str, List[dict]], *args, **kwargs):
        raise NotImplementedError("ChatAgent 子类必须实现 generate 方法")
    
    async def async_generate(self, prompt: Union[str, List[dict]], *args, **kwargs):
        loop = asyncio.get_running_loop()
        for block in await self.run_in_executor(self.generate, prompt, *args, **kwargs):
            yield block

    def _fetch_final_output(self, output_text, chat_memory):
        # 提取最终答案
        self._final_answer = extract_text(output_text, self.fetching_final_answer)

        # 提取上下文
        context = extract_text(output_text, self.fetching_context)

        final_output_text = extract_text(output_text, (self.start_marker, self.end_marker))
        # 保存到最近输出
        self._last_output = final_output_text

        messages = Messages([("ai", final_output_text)], style=self.style).to_list()
        # 追加到短期记忆中
        chat_memory.extend(messages)
        # 追加到长期记忆中
        self.remember_response(messages)

        return {
            "final_output_text": final_output_text,
            "context": context
        }

    def _handle_tool_calls(self, final_output_text, chat_memory, tools_behavior: str="none"):
        if "parse" in tools_behavior:
            for handler in self.tools_handlers:
                steps = handler.extract_tools_call(final_output_text)
                if steps and "execute" in tools_behavior:
                    for block in handler.handle(
                        steps,
                        short_term_memory=chat_memory,
                        long_term_memory=self.memory
                    ):
                        yield block

    def _patch_knowledge(self, messages: Messages):
        """
        根据当前的记忆和知识，对提示语进行补充。

        补充内容包括：
        - 默认保存在 __DOCS__ 目录中的已有知识
        - 默认保存在临时目录中的经验信息
        - 已经添加的资源信息
        """
        kg = ""
        existing_text = "\n".join([m['content'] for m in Messages(messages).to_list(style="text")])
        if self.knowledge:
            for item in self.get_knowledge(self.task):
                if item not in existing_text:
                    kg += item
        patch_info = ""
        if kg:
            patch_info += f"回答时请参考已有知识：\n@knowledge\n{kg}\n"
        if patch_info:
            add_messages = Messages([
                ("user", patch_info),
                ("assistant", "ok")
            ], style=self.style).to_list()

            self.memory.insert(-1, add_messages[0])
            self.memory.insert(-1, add_messages[1])

            messages.insert(-1, add_messages[0])
            messages.insert(-1, add_messages[1])

        return messages

    def call(self, prompt: Any, tools_behavior: str=None, **kwargs):
        """
        执行对话生成，并处理工具调用。

        illufly 的默认设计中有很多便利，比如自动执行工具回调。

        但这种默认设计有时也带来额外的困扰，例如，当你在工具回调之后想谈论刚刚的工具回调过程，而不是想调用它。
        此时，你可以使用 tools_behavior 参数临时修改对待工具回调的行为。
        """

        # 兼容 Runnable 类型，将其上一次的输出作为 prompt 输入
        if not isinstance(prompt, PromptTemplate) and isinstance(prompt, Runnable):
            prompt = prompt.selected.last_output

        new_chat = kwargs.pop("new_chat", False)

        messages_std = Messages(prompt, style="text")
        self._task = messages_std.to_list()[-1]['content']
        yield EventBlock("user", self._task)

        # 根据模板中是否直接使用 tools_desc 来替换 tools 参数
        self._tools_to_exec = self.get_tools(kwargs.get("tools", []))
        for tool in self._tools_to_exec:
            self.bind_consumer(tool, dynamic=True)
        if "tools_desc" in self.get_bound_vars(messages_std, new_chat=new_chat):
            kwargs["tools"] = None
        else:
            kwargs["tools"] = self.get_tools_desc(kwargs.get("tools", [])) or None

        # 重新绑定工具处理的 handlers
        for h in self.tools_handlers:
            h.reset(self._tools_to_exec)

        remember_rounds = kwargs.pop("remember_rounds", self.remember_rounds)
        yield EventBlock("info", f'记住 {remember_rounds} 轮对话')

        chat_memory = self.build_chat_memory(
            prompt=prompt, # 这里依然使用 prompt
            new_chat=new_chat,
            remember_rounds=remember_rounds
        )
        chat_memory = self._patch_knowledge(chat_memory)
        if self.recalled_knowledge:
            kg_source = {}
            for item in self.recalled_knowledge:
                if isinstance(item, Document):
                    src = item.meta.get("source", "无来源文档")
                else:
                    src = "直接资料"
                if src not in kg_source:
                    kg_source[src] = []
                kg_source[src].append(item)
            for k, v in kg_source.items():
                yield EventBlock("RAG", f"{k}：发现 {len(v)} 条资料")

        to_continue_call_llm = True
        while to_continue_call_llm:
            to_continue_call_llm = False

            # 执行模型生成任务
            output_text, tools_call = "", []
            content_id = str(uuid.uuid1().hex)

            for block in self.generate(chat_memory, **kwargs):
                block.content_id = content_id
                yield block

                if block.block_type in ["chunk", "text"]:
                    output_text += block.text
                elif block.block_type == "final_text":
                    output_text = block.text
                elif block.block_type == "tools_call_chunk":
                    tools_call.append(json.loads(block.text))

            openai_tools_calling_steps = merge_tool_calls(tools_call)
            if openai_tools_calling_steps:
                # 从返回参数中解析工具
                handler_openai = OpenAIToolsCalling(tools_to_exec=self._tools_to_exec)
                # 处理在返回结构中包含的 openai 风格的 tools-calling 工具调用，包括将结果追加到记忆中
                for block in handler_openai.handle(openai_tools_calling_steps, chat_memory, self.memory):
                    block.content_id = content_id
                    if block.block_type == "chunk":
                        block.block_type = "tool_resp_chunk"
                    elif block.block_type == "text":
                        block.block_type = "tool_resp_text"
                    yield block

                    if isinstance(block, EventBlock) and block.block_type == "final_tool_resp":
                        to_continue_call_llm = True
            else:
                final_output_text = self._fetch_final_output(output_text, chat_memory)
                yield EventBlock("final_text", final_output_text["final_output_text"], content_id=content_id)
                if final_output_text["context"]:
                    yield EventBlock("context", final_output_text["context"], content_id=content_id)

                _tools_behavior = tools_behavior or self.tools_behavior
                to_continue_call_llm = False
                for block in self._handle_tool_calls(final_output_text["final_output_text"], chat_memory, _tools_behavior):
                    block.content_id = content_id
                    if block.block_type == "chunk":
                        block.block_type = "tool_resp_chunk"
                    elif block.block_type == "text":
                        block.block_type = "tool_resp_text"
                    yield block

                    if isinstance(block, EventBlock) and block.block_type == "final_tool_resp":
                        if "continue" in _tools_behavior:
                            to_continue_call_llm = True

        if self.end_chk:
            yield EndBlock(self.last_output)
        self.save_memory()

    async def async_call(self, prompt: Any, tools_behavior: str=None, **kwargs):
        # 兼容 Runnable 类型，将其上一次的输出作为 prompt 输入
        if not isinstance(prompt, PromptTemplate) and isinstance(prompt, Runnable):
            prompt = prompt.selected.last_output

        new_chat = kwargs.pop("new_chat", False)

        messages_std = Messages(prompt, style="text")
        self._task = messages_std.to_list()[-1]['content']
        yield EventBlock("user", self._task)

        # 根据模板中是否直接使用 tools_desc 来替换 tools 参数
        self._tools_to_exec = self.get_tools(kwargs.get("tools", []))
        for tool in self._tools_to_exec:
            self.bind_consumer(tool, dynamic=True)
        if "tools_desc" in self.get_bound_vars(messages_std, new_chat=new_chat):
            kwargs["tools"] = None
        else:
            kwargs["tools"] = self.get_tools_desc(kwargs.get("tools", []))

        # 重新绑定工具处理的 handlers
        for h in self.tools_handlers:
            h.reset(self._tools_to_exec)

        remember_rounds = kwargs.pop("remember_rounds", self.remember_rounds)
        yield EventBlock("info", f'记住 {remember_rounds} 轮对话')

        chat_memory = self.build_chat_memory(
            prompt=prompt,
            new_chat=new_chat,
            remember_rounds=remember_rounds
        )
        chat_memory = self._patch_knowledge(chat_memory)

        to_continue_call_llm = True
        while to_continue_call_llm:
            to_continue_call_llm = False

            output_text, tools_call = "", []
            content_id = str(uuid.uuid1().hex)

            async for block in self.async_generate(chat_memory, **kwargs):
                block.content_id = content_id
                yield block

                if block.block_type in ["chunk", "text"]:
                    output_text += block.text
                elif block.block_type == "final_text":
                    output_text = block.text
                elif block.block_type == "tools_call_chunk":
                    tools_call.append(json.loads(block.text))

            openai_tools_calling_steps = merge_tool_calls(tools_call)
            if openai_tools_calling_steps:
                # 从返回参数中解析工具
                handler_openai = OpenAIToolsCalling(tools_to_exec=self._tools_to_exec)
                # 处理在返回结构中包含的 openai 风格的 tools-calling 工具调用，包括将结果追加到记忆中
                async for block in handler_openai.async_handle(openai_tools_calling_steps, chat_memory, self.memory):
                    block.content_id = content_id
                    if block.block_type == "chunk":
                        block.block_type = "tool_resp_chunk"
                    elif block.block_type == "text":
                        block.block_type = "tool_resp_text"
                    yield block

                    if isinstance(block, EventBlock) and block.block_type == "final_tool_resp":
                        to_continue_call_llm = True

            else:
                final_output_text = self._fetch_final_output(output_text, chat_memory)
                yield EventBlock("final_text", final_output_text["final_output_text"], content_id=content_id)
                if final_output_text["context"]:
                    yield EventBlock("context", final_output_text["context"], content_id=content_id)

                _tools_behavior = tools_behavior or self.tools_behavior
                to_continue_call_llm = False
                for block in self._handle_tool_calls(final_output_text["final_output_text"], chat_memory, _tools_behavior):
                    block.content_id = content_id
                    if block.block_type == "chunk":
                        block.block_type = "tool_resp_chunk"
                    elif block.block_type == "text":
                        block.block_type = "tool_resp_text"
                    yield block

                    if isinstance(block, EventBlock) and block.block_type == "final_tool_resp":
                        if "continue" in _tools_behavior:
                            to_continue_call_llm = True

        if self.end_chk:
            yield EndBlock(self.last_output)
        self.save_memory()



from typing import List, Union, Optional, Callable, Dict, Generator, AsyncGenerator
from ...utils import raise_invalid_params
from ...io import EventBlock
from .base import Runnable
import inspect
import random

def select_first(consumer_dict: Dict, runnables: List[Runnable]):
    return runnables[0]

def select_random(consumer_dict: Dict, runnables: List[Runnable]):
    """
    从runnables列表中随机选择一个Runnable对象
    """
    if not runnables:
        raise ValueError("runnables列表不能为空")
    return random.choice(runnables)

class End(Runnable):
    @classmethod
    def allowed_params(cls):
        return {
            **Runnable.allowed_params()
        }

    def __init__(self,**kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        kwargs.update({"name": "__End__"})
        super().__init__(**kwargs)

        self.description = "我是一个结束标志"

    def call(self, *args, **kwargs):
        pass

class Selector(Runnable):
    """
    路由选择 Runnable 对象的智能体，并将任务分发给被选择对象执行。

    可以根据模型，以及配置模型所需的工具集、资源、数据、handlers等不同参数，构建为不同的智能体对象。

    Selector 需要执行 select 方法来确定选中对象，然后你可以通过 selected 属性来提取选中对象。
    但如果你使用 Selector 的 call 方法，则会自动调用一次 select 方法。

    由于 select 方法不是一个幂等操作，可能两次 select 方法的返回结果并不相同，因此你应当非常小心地管理 select 方法的调用。
    """
    @classmethod
    def allowed_params(cls):
        return {
            "runnables": "参与路由的 Runnable 列表",
            "condition": "选择条件，可以是自定义 Callable 或通过字符串选择内置的条件函数[first, random, similar]。自定义时有三种参数：无参数、仅 consumer_dict 一个参数 或同时提供 consumer_dict 和 runnables 两个参数",
            "embeddings": "如果 condition 是 similar 则需要提供用于相似度计算的 embeddings 实例",
            **Runnable.allowed_params()
        }

    def __init__(
        self,
        condition: Union[Callable, str] = None,
        runnables: List[Runnable] = None,
        embeddings: "BaseEmbeddings" = None,
        **kwargs
    ):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        super().__init__(**kwargs)

        if runnables:
            self.runnables = runnables if isinstance(runnables, list) else [runnables]
        else:
            self.runnables = []

        if runnables and not all(isinstance(router, Runnable) for router in self.runnables):
            raise ValueError("param runnables must be a list of Runnables")

        self.condition = self.get_condition(condition)
        self.embeddings = embeddings

        for runnable in self.runnables:
            self.bind_consumer(runnable)
        
        self._selected = None

    def get_condition(self, condition):
        default_selected = {
            "first": select_first,
            "random": select_random,
            "similar": self.similar_with_description
        }
        if condition is None:
            return default_selected['first']
        elif isinstance(condition, str):
            return default_selected.get(condition.lower(), default_selected['first'])
        elif isinstance(condition, Callable):
            return condition
        else:
            raise ValueError("param condition must be a Callable")

    @property
    def selected(self):
        if self._selected is None:
            self.select()
        return self._selected

    def select(self):
        """
        selected 可以是一个 Runnable 对象，也可以是一个字符串（表示备选 Runnable 的名称）
        """
        signature = inspect.signature(self.condition)
        if len(signature.parameters) == 0:
            resp = self.condition()
        elif len(signature.parameters) == 1:
            resp = self.condition(self.consumer_dict)
        else:
            resp = self.condition(self.consumer_dict, self.runnables)

        if isinstance(resp, str):
            if resp.lower() == "end":
                self._selected = End()
        self._selected = resp
        return self._selected

    def call(self, *args, **kwargs) -> List[dict]:
        self.select()
        yield from self.selected.call(*args, **kwargs)
        self._last_output = self.selected.last_output

    async def async_call(self, *args, **kwargs) -> List[dict]:
        self.select()
        resp = await self.selected.async_call(*args, **kwargs)
        if isinstance(resp, Generator):
            for block in resp:
                yield block
        elif isinstance(resp, AsyncGenerator):
            async for block in resp:
                yield block
        else:
            yield resp
        self._last_output = resp.last_output

    def similar_with_description(self):
        if not self.embeddings:
            raise ValueError("embeddings is not set")

        from ...community.faiss import FaissDB

        db = FaissDB(embeddings=self.embeddings)
        for run in self.runnables:
            if run.description:
                db.load_text(run.description)

        query = self.consumer_dict.get("task")
        if query:
            results = db(query)
            desc = results[0]
            for run in self.runnables:
                if run.description == desc.text:
                    return run
            raise ValueError(f"router {desc.text} not found in {self.runnables}")
        else:
            return self.runnables[0]



from typing import Union, List, Optional
from urllib.parse import urlparse
import os

from ..http import EventBlock, save_resource
from ...types import BaseAgent
from ...utils import raise_invalid_params

class CogView(BaseAgent):
    """    
    [CogView API](https://open.bigmodel.cn/dev/api/image-model/cogview)
    """
    @classmethod
    def allowed_params(cls):
        return {
            "model": "模型名称",
            "api_key": "API_KEY",
            "base_url": "API_BASE_URL",
            **BaseAgent.allowed_params()
        }

    def __init__(self, model: str=None, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())
        try:
            from zhipuai import ZhipuAI
        except ImportError:
            raise ImportError(
                "Could not import zhipuai package. "
                "Please install it via 'pip install -U zhipuai'"
            )

        super().__init__(threads_group="COGVIEW", **kwargs)

        self.default_call_args = {
            "model": model or "cogview-3-plus"
        }
        self.model_args = {
            "api_key": kwargs.get("api_key", os.getenv("ZHIPUAI_API_KEY")),
            "base_url": kwargs.get("base_url", os.getenv("ZHIPUAI_BASE_URL")),
        }
        self.client = ZhipuAI(**self.model_args)

        self.description = "我擅长根据你的文字提示描述生成图片，你必须在 prompt 中详细描述生成要求，你必须展开描述，比如镜头、光线、细部等。"
        self.tool_params = {
            "prompt": "请尽量详细描述生成要求的细节",
            "size": "图片尺寸, 可选范围: [1024x1024,768x1344,864x1152,1344x768,1152x864,1440x720,720x1440]，默认是1024x1024。",
            "output": "指定生成图片的名称，应当包括扩展名"
        }

    def call(
        self, 
        prompt: str=None,
        size: str=None,
        output: Optional[Union[str, List[str]]] = None,
        **kwargs
    ):
        _kwargs = self.default_call_args
        _kwargs.update({
            "prompt": prompt,
            "size": size or "1024x1024",
            **kwargs,
        })
        if isinstance(output, str):
            output = [output]

        resp = self.client.images.generations(**_kwargs)
        for result_index, result in enumerate(resp.data):
            url = result.url
            yield EventBlock("image_url", url)
            if not output or result_index >= len(output):
                parsed_url = urlparse(url)
                filename = os.path.basename(parsed_url.path)
                output_path = f"{filename.rsplit('.', 1)[0]}.{filename.rsplit('.', 1)[1]}"
            else:
                output_path = output[result_index]
            yield from save_resource(url, output_path)



from typing import Union, List, Optional, Dict, Any

from ...io import EventBlock, NewLineBlock
from ...types import ChatAgent
from ...utils import raise_invalid_params
from ..http import (
    EventBlock,

    get_headers,
    validate_output_path,

    send_request,
    check_task_status,
    save_resource,

    async_send_request,
    async_check_task_status,
    async_save_resource,

    confirm_base64_or_uri
)

import os
import json

class ChatZhipu(ChatAgent):
    @classmethod
    def allowed_params(cls):
        return {
            "model": "模型名称",
            **ChatAgent.allowed_params()
        }

    def __init__(self, model: str=None, extra_args: dict={}, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())
        try:
            from zhipuai import ZhipuAI
        except ImportError:
            raise ImportError(
                "Could not import zhipuai package. "
                "Please install it via 'pip install -U zhipuai'"
            )

        super().__init__(threads_group="CHAT_ZHIPU", **kwargs)

        self.default_call_args = {
            "model": model or "glm-4-flash"
        }
        self.model_args = {
            "api_key": kwargs.get("api_key", os.getenv("ZHIPUAI_API_KEY")),
            "base_url": kwargs.get("base_url", os.getenv("ZHIPUAI_BASE_URL")),
            **extra_args
        }
        self.client = ZhipuAI(**self.model_args)

    def generate(
        self,
        messages: List[dict],
        **kwargs
    ):
        _kwargs = self.default_call_args
        _kwargs.update({
            "messages": messages,
            **kwargs,
            **{"stream": True}
        })

        completion = self.client.chat.completions.create(**_kwargs)

        usage = {}
        output = []
        request_id = None
        for response in completion:
            if response.usage:
                request_id = response.id
                usage = response.usage
            if response.choices:
                ai_output = response.choices[0].delta
                if ai_output.tool_calls:
                    for func in ai_output.tool_calls:
                        func_json = {
                            "id": func.id or "",
                            "type": func.type or "function",
                            "function": {
                                "name": func.function.name or "",
                                "arguments": func.function.arguments or ""
                            }
                        }
                        output.append({"tools_call_chunk": func_json})
                        yield EventBlock("tools_call_chunk", json.dumps(func_json, ensure_ascii=False))
                else:
                    content = ai_output.content
                    if content:
                        output.append({"chunk": content})
                        yield EventBlock("chunk", content)
        yield NewLineBlock()
        usage_dict = {
            "prompt_tokens": usage.prompt_tokens if usage else None,
            "completion_tokens": usage.completion_tokens if usage else None,
            "total_tokens": usage.total_tokens if usage else None
        }
        yield EventBlock(
            "usage",
            json.dumps(usage_dict, ensure_ascii=False),
            calling_info={
                "request_id": request_id,
                "input": _kwargs,
                "output": output,
            }
        )

class ChatZhipuVL(ChatZhipu):
    def __init__(self, model: str=None, **kwargs):
        super().__init__(model=(model or "glm-4v-plus"), style="openai_vl", **kwargs)

    def generate(self, messages: List[dict], **kwargs):
        for m in messages:
            if isinstance(m['content'], list):
                for obj in m['content']:
                    if isinstance(obj, dict):
                        if obj['type'] == "image_url":
                            obj['image_url']['url'] = confirm_base64_or_uri(obj['image_url']['url'])
                        elif obj['type'] == "video_url":
                            obj['video_url']['url'] = confirm_base64_or_uri(obj['video_url']['url'])
        _kwargs = self.default_call_args
        _kwargs.update({
            "messages": messages,
            **kwargs,
            **{"stream": True}
        })

        completion = self.client.chat.completions.create(**_kwargs)

        usage = {}
        output = []
        request_id = None
        for response in completion:
            if response.usage:
                request_id = response.id
                usage = response.usage
            if response.choices:
                ai_output = response.choices[0].delta
                content = ai_output.content
                if content:
                    output.append({"chunk": content})
                    yield EventBlock("chunk", content)
        yield NewLineBlock()
        usage_dict = {
            "prompt_tokens": usage.prompt_tokens if usage else None,
            "completion_tokens": usage.completion_tokens if usage else None,
            "total_tokens": usage.total_tokens if usage else None
        }
        yield EventBlock(
            "usage",
            json.dumps(usage_dict, ensure_ascii=False),
            calling_info={
                "request_id": request_id,
                "input": _kwargs,
                "output": output,
            }
        )



import json
from typing import AsyncIterable, Union
import logging
import asyncio

from .block import EventBlock

async def event_stream(block, verbose: bool=False, **kwargs):
    """
    生成适合于 Web 处理的 SSE 事件流格式的数据。
    """

    if isinstance(block, EventBlock):
        yield f"event: {block.block_type}\ndata: {block.text}\n\n"
    elif isinstance(block, str):
        yield f"data: {block}\n\n"
    else:
        # logging.error(f"Unknown block type: {block}")
        # 在生产环境中，可能不想中断整个流，而是跳过无效的块
        pass
    await asyncio.sleep(0)



from typing import Any, List
from http import HTTPStatus

import os
from ...types import BaseEmbeddings
from ...utils import raise_invalid_params

class ZhipuEmbeddings(BaseEmbeddings):
    """支持最新的阿里云模型服务灵积API的文本向量模型"""

    @classmethod
    def allowed_params(cls):
        return {
            "model": "模型名称",
            "api_key": "API_KEY",
            "dim": "编码时使用的向量维度",
            **BaseEmbeddings.allowed_params()
        }

    def __init__(self, model: str=None, api_key: str=None, dim: int=None, extra_args: dict={}, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        super().__init__(
            model=model or "embedding-3",
            api_key=api_key or os.getenv("ZHIPUAI_API_KEY"),
            dim=dim or 2048,
            **kwargs
        )

        try:
            from zhipuai import ZhipuAI
            self.client = ZhipuAI(api_key=api_key, **extra_args)
        except ImportError:
            raise ImportError(
                "Could not import zhipuai package. "
                "Please install it via 'pip install -U zhipuai'"
            )

    def query(self, text: str, **kwargs) -> List[float]:
        """
        查询文本向量。
        """
        return self.embed_documents([text], **kwargs)[0]

    def embed_documents(self, texts: List[str], **kwargs) -> List[List[float]]:
        """
        编码文本向量。
        """
        response = self.client.embeddings.create(
            model=self.model,
            input=texts,
            dimensions=self.dim,
            **kwargs
        )
        return [ed.embedding for ed in response.data]




from typing import Union, List, Dict, Any
import os
import time
import requests
import aiohttp
import asyncio
import hashlib
import base64

from ..io import EventBlock
from ..config import get_env

CHECK_RESULT_SECONDS = get_env("HTTP_CHECK_RESULT_SECONDS")
DASHSCOPE_BASE_URL = get_env("DASHSCOPE_BASE_URL")
ZHIPUAI_API_TOOLS_BASE_URL = get_env("ZHIPUAI_API_TOOLS_BASE_URL")

# 保留当日上传过的缓存
# 缓存的命名：以年月作为子目录，将上传内容和日期一起做MD5编码作为文件名称

def get_upload_cache(file_path: str, upload_server: str="aliyun_oss") -> str:
    UPLOAD_CACHE_DIR = get_env("ILLUFLY_UPLOAD_CACHE")
    date_dir = time.strftime("%Y-%m")
    cache_dir = os.path.join(UPLOAD_CACHE_DIR, date_dir, upload_server)
    if os.path.exists(file_path):
        with open(file_path, "rb") as f:
            file_content = f.read()
            content_md5 = hashlib.md5(file_content).hexdigest() + ".json"
            cache_path = os.path.join(cache_dir, content_md5)
            if os.path.exists(cache_path):
                with open(cache_path, "r") as f:
                    return {"upload_url": f.read(), "cache_path": cache_path}
            else:
                return {"upload_url": None, "cache_path": cache_path}
    return {"upload_url": None, "cache_path": None}

def update_upload_cache(file_path: str, upload_url: str, upload_server: str="aliyun_oss") -> dict:
    cache = get_upload_cache(file_path, upload_server)
    cache_file_path = cache['cache_path']

    if cache_file_path and upload_url:
        os.makedirs(os.path.dirname(cache_file_path), exist_ok=True)
        with open(cache_file_path, 'w') as f:
            f.write(upload_url)
            return {"upload_url": upload_url, "cache_path": cache_file_path}

    return {"upload_url": None, "cache_path": cache_file_path}

def confirm_upload_file(model: str=None, uri: str=None, api_key: str=None):
    try:
        from dashscope.utils.oss_utils import upload_file
    except ImportError:
        raise RuntimeError(
            "Could not import dashscope package. "
            "Please install it via 'pip install -U dashscope'"
        )

    if uri and (uri.startswith('http://') or uri.startswith('https://') or uri.startswith('oss://')):
        return uri
    
    media_file_path = os.path.join(get_env("ILLUFLY_RESOURCE"), uri)
    if os.path.exists(media_file_path):
        cache = get_upload_cache(media_file_path, "aliyun_oss")
        if cache['cache_path'] and cache['upload_url']:
            return cache['upload_url']
        else:
            oss_file = f"file://{media_file_path}"
            upload_url = upload_file(model, oss_file, api_key)
            update_upload_cache(media_file_path, upload_url, "aliyun_oss")
            return upload_url
    else:
        raise ValueError(f"Invalid URI path: {uri}")

def confirm_base64_or_uri(uri: str) -> str:
    if uri:
        if uri.startswith('http://') or uri.startswith('https://') or uri.startswith('oss://'):
            return uri
        else:
            with open(uri, 'rb') as file:
                base64_encoded = base64.b64encode(file.read()).decode('utf-8')
                return base64_encoded
    else:
        return uri

def get_headers(api_key: str, enable_async: bool = True) -> Dict[str, str]:
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {api_key}",
        "X-DashScope-OssResourceResolve": "enable"
    }
    if enable_async:
        headers["X-DashScope-Async"] = "enable"
    return headers

def validate_output_path(output_path: Union[str, List[str]], n: int) -> List[str]:
    if isinstance(output_path, str):
        output_path = [output_path]
    if output_path and len(output_path) != n:
        raise ValueError(f"Invalid output_path: {output_path}, please ensure the number of images is consistent with the n value")
    return output_path

def save_resource(url: str, path: str):
    res_to_save = os.path.join(get_env("ILLUFLY_RESOURCE"), path)
    if res_to_save and not os.path.exists(res_to_save):
        os.makedirs(os.path.dirname(res_to_save), exist_ok=True)
    with open(res_to_save, 'wb+') as f:
        f.write(requests.get(url).content)
        yield EventBlock("chunk", f'resource was generated and saved to {path}; ')

async def async_save_resource(url: str, path: str):
    res_to_save = os.path.join(get_env("ILLUFLY_RESOURCE"), path)
    if res_to_save and not os.path.exists(res_to_save):
        os.makedirs(os.path.dirname(res_to_save), exist_ok=True)
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            content = await response.read()
            with open(res_to_save, 'wb+') as f:
                f.write(content)
                yield EventBlock("chunk", f'an resource was generated and saved to {path}')

def get_request(url: str, headers: Dict[str, str]) -> Dict[str, Any]:
    response = requests.get(url, headers=headers)
    return response.json()

async def async_get_request(url: str, headers: Dict[str, str]) -> Dict[str, Any]:
    async with aiohttp.ClientSession() as session:
        async with session.get(url, headers=headers) as response:
            return await response.json()

def send_request(url: str, headers: Dict[str, str], data: Dict[str, Any]) -> Dict[str, Any]:
    response = requests.post(url, headers=headers, json=data)
    return response.json()

async def async_send_request(url: str, headers: Dict[str, str], data: Dict[str, Any]) -> Dict[str, Any]:
    async with aiohttp.ClientSession() as session:
        async with session.post(url, headers=headers, json=data) as response:
            return await response.json()

def check_task_status(status_result: Dict[str, Any], task_id: str, headers: Dict[str, str]) -> Dict[str, Any]:
    while True:
        time.sleep(CHECK_RESULT_SECONDS)
        status_response = requests.get(f"{DASHSCOPE_BASE_URL}/tasks/{task_id}", headers=headers)
        _status_result = status_response.json()
        yield EventBlock("info", f'{task_id}: {_status_result["output"]["task_status"]}')

        if _status_result['output']['task_status'] in ['SUCCEEDED', 'FAILED']:
            status_result.update(_status_result)
            break

async def async_check_task_status(status_result: Dict[str, Any], task_id: str, headers: Dict[str, str]) -> Dict[str, Any]:
    async with aiohttp.ClientSession() as session:
        while True:
            await asyncio.sleep(CHECK_RESULT_SECONDS)
            async with session.get(f"{DASHSCOPE_BASE_URL}/tasks/{task_id}", headers=headers) as status_response:
                _status_result = await status_response.json()
                yield EventBlock("info", f'{task_id}: {_status_result["output"]["task_status"]}')
                if _status_result['output']['task_status'] in ['SUCCEEDED', 'FAILED']:
                    status_result.update(_status_result)
                    break





from typing import Union, List, Optional, Dict, Any

from ...io import EventBlock, NewLineBlock
from ...types import ChatAgent
from ...utils import raise_invalid_params

import os
import json

class ChatOpenAI(ChatAgent):
    @classmethod
    def allowed_params(cls):
        return {
            "model": "模型名称",
            "imitator": "兼容 OpenAI 接口协议的模型来源，默认 imitator='OPENAI'，即从环境变量中读取 OPENAI_API_KEY 和 OPENAI_BASE_URL。",
            **ChatAgent.allowed_params()
        }

    def __init__(self, model: str=None, imitator: str=None, extra_args: dict={}, **kwargs):
        """
        使用 imitator 参数指定兼容 OpenAI 接口协议的模型来源，默认 imitator="OPENAI"。
        只需要在环境变量中配置 imitator 对应的 API_KEY 和 BASE_URL 即可。
        
        例如：
        QWEN_API_KEY=sk-...
        QWEN_BASE_URL=https://dashscope.aliyuncs.com/compatible-mode/v1

        然后使用类似 `ChatOpenAI(imitator="QWEN")` 的代码就可以使用千问系列模型。
        """
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        try:
            from openai import OpenAI
        except ImportError:
            raise ImportError(
                "Could not import openai package. "
                "Please install it via 'pip install -U openai'"
            )

        imitator = (imitator or "").upper() or "OPENAI"
        super().__init__(threads_group=f"CHAT_{imitator}", **kwargs)

        self.default_call_args = {
            "model": model or os.getenv(f"{imitator}_MODEL_ID") or "gpt-3.5-turbo"
        }
        self.model_args = {
            "base_url": kwargs.pop("base_url", os.getenv(f"{imitator}_BASE_URL")),
            "api_key": kwargs.pop("api_key", os.getenv(f"{imitator}_API_KEY")),
            **extra_args
        }
        self.client = OpenAI(**self.model_args)

    def generate(
        self,
        messages: List[dict],
        **kwargs
    ):
        from openai import OpenAI

        _kwargs = self.default_call_args
        _kwargs.update({
            "messages": messages,
            **kwargs,
            **{"stream": True, "stream_options": {"include_usage": True}}
        })

        completion = self.client.chat.completions.create(**_kwargs)

        usage = {}
        output = []
        request_id = None
        for response in completion:
            if response.usage:
                usage = response.usage
            if response.choices:
                ai_output = response.choices[0].delta
                if ai_output.tool_calls:
                    for func in ai_output.tool_calls:
                        func_json = {
                            # "index": func.index or 0,
                            "id": func.id or "",
                            "type": func.type or "function",
                            "function": {
                                "name": func.function.name or "",
                                "arguments": func.function.arguments or ""
                            }
                        }
                        output.append({"tools_call_chunk": func_json})
                        yield EventBlock("tools_call_chunk", json.dumps(func_json, ensure_ascii=False))
                else:
                    content = ai_output.content
                    if content:
                        output.append({"chunk": content})
                        yield EventBlock("chunk", content)
        yield NewLineBlock()
        usage_dict = {
            "prompt_tokens": usage.prompt_tokens if usage else None,
            "completion_tokens": usage.completion_tokens if usage else None,
            "total_tokens": usage.total_tokens if usage else None
        }
        yield EventBlock(
            "usage",
            json.dumps(usage_dict, ensure_ascii=False),
            calling_info={
                "request_id": request_id,
                "input": _kwargs,
                "output": output,
            }
        )



import re
import os
import json
import hashlib
import tiktoken
from typing import List, Union, Dict, Any, Tuple
from .config import get_env

def filter_kwargs(kwargs: Dict, allowed_params: Dict):
    """
    根据 allowed_params 过滤掉不支持的参数。
    """
    return {k: v for k, v in kwargs.items() if k in allowed_params}

def raise_invalid_params(kwargs: Dict, allowed_params: Dict):
    """
    根据 allowed_params 检查参数是否合法。
    """
    invalid_params = [k for k in kwargs.keys() if k not in allowed_params]
    if invalid_params:
        raise ValueError(f"invalid parameters: {invalid_params}")

def raise_not_install(packages):
    """
    如果指定的包未安装，则抛出错误。
    """
    print(f"please install package: '{packages}' with pip or poetry")
    # auto install package
    # subprocess.check_call([sys.executable, "-m", "pip", "install", package_name])

def raise_not_supply_all(info: str, *args):
    """
    如果所有参数都为None，则抛出错误。
    """
    if all(arg is None for arg in args):
        raise ValueError(info)

def extract_segments(text: str, start_marker: str, end_marker: str, mode: str = 'all', include_markers: bool = False) -> List[str]:
    """
    根据模式提取文本中符合条件的片段。
    mode='all'：提取每一对start_marker和end_marker之间的内容。
    mode='first_last'：提取第一个start_marker和最后一个end_marker之间的内容。
    """
    lines = text.split('\n')
    segments = []
    capture = False
    current_segment = []

    if mode == 'all':
        for line in lines:
            if not capture and line.startswith(start_marker):
                capture = True
                current_segment = [] if not include_markers else [line]
            elif capture and line.startswith(end_marker):
                if include_markers:
                    current_segment.append(line)
                segments.append('\n'.join(current_segment))
                capture = False
            elif capture:
                current_segment.append(line)

    else:
        start_index = None
        end_index = None

        for i, line in enumerate(lines):
            if not capture and line.startswith(start_marker) and start_index is None:
                start_index = i if include_markers else i + 1
            if capture and line.startswith(end_marker):
                end_index = i if include_markers else i + 1

        segments.append('\n'.join(lines[start_index:end_index]))

    return segments

def extract_text(resp_md: str, marker: Tuple[str, str]):
    """
    如果指定开始和结束的标记，就提取标记中间的文本，并移除标记所在的行。
    一旦文本出现Markdown标题（若干个#开头的行），之后的内容就都不要进行start_marker匹配。
    """
    start_marker, end_marker = marker

    if start_marker and end_marker:
        start_lines = resp_md.split('\n')
        # 查找第一个Markdown标题的索引
        markdown_title_index = next((i for i, line in enumerate(start_lines) if line.strip().startswith('#')), len(start_lines))
        
        # 在第一个Markdown标题之前查找start_marker
        start_index = next((i for i, line in enumerate(start_lines[:markdown_title_index]) if re.search(start_marker, line, re.IGNORECASE)), None)
        end_index = next((i for i, line in enumerate(reversed(start_lines), 1) if re.search(end_marker, line, re.IGNORECASE)), None)

        if start_index is not None and end_index is not None and start_index < len(start_lines) - end_index:
            extracted_text = '\n'.join(start_lines[start_index+1:len(start_lines)-end_index]).strip()
            return extracted_text

    return resp_md

def extract_final_answer(text: str, final_answer_prompt: str="最终答案"):
    """
    提取文本中以 final_answer_prompt 开头的文本。
    """
    final_answer_lines = []
    capture = False

    # 预处理 final_answer_prompt，仅保留字母、中文、数字、下划线和横杠
    processed_prompt = re.sub(r'[^\w\u4e00-\u9fff-]', '', final_answer_prompt)

    for line in text.split('\n'):
        # 预处理行，仅保留字母、中文、数字、下划线和横杠
        processed_line = re.sub(r'[^\w\u4e00-\u9fff-]', '', line)
        if processed_line == processed_prompt:
            capture = True
            # 直接添加整行，而不是从 final_answer_prompt 的长度位置开始
            final_answer_lines.append(line.strip())
        elif capture:
            final_answer_lines.append(line)
    final_answer_text = '\n'.join(final_answer_lines).strip()

    return extract_text(final_answer_text, ("```", "```")) if final_answer_text else ''

def hash_text(text):
    """
    计算文本的MD5哈希值。主要用于生成反向索引，便于缓存命中等。
    """
    text_bytes = text.encode('utf-8')
    hash_object = hashlib.md5(text_bytes)
    return hash_object.hexdigest()

def clean_filename(filename: str):
    """
    先将除字母、数字、中文、下划线和短横线之外的字符替换为下划线;
    再将多个连续的下划线或短横线替换为单个下划线。
    """
    cleaned_filename = re.sub(r'[^\w\s-]', '_', filename)
    cleaned_filename = re.sub(r'[-_ ]+', '_', cleaned_filename)
    return cleaned_filename

def safety_path(path: str):
    """
    确保路径中不包含 .. ，防止通过路径注入获得系统中其他资源的路径。
    替换路径中的 .. 为 .，并返回安全过滤后的路径。
    """
    return os.path.normpath(re.sub(r"\.\.+", ".", path)) if path else ''

def minify_text(text: str, limit: int=100) -> str:
    """
    在长度限制下，仅保留文本的开头部份即可，一般用于调试信息。
    压缩文本，剔除左右两侧的空格和换行，仅保留第一个换行之前的文字，超出后limit后用省略号代替。
    """
    raw_len = len(text)
    if not text:
        return ''
    
    text = text.strip()
    minified_text = text[:limit].replace("\n", "<br>")
    if len(minified_text) < raw_len:
        minified_text += f'...'
    
    return minified_text

def compress_text(text: str, start_limit: int=100, end_limit: int=100, delta: int=50) -> str:
    """
    在长度限制下，保留尽量多的文本，一般用于提示语参考。
    压缩文本，如果文本长度超过指定限制，则只保留前后部分，并用省略号连接。
    """
    if not text:
        return ''

    if len(text) <= start_limit + end_limit + delta:
        # 如果文本长度小于或等于前后限制之和，则直接返回原文本
        return text
    else:
        # 否则，保留前后部分并用省略号连接
        return text[:start_limit] + f"\n...(省略{len(text)-start_limit-end_limit}字)\n" + text[-end_limit:]

def merge_tool_calls(blocks: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    合并流式输出的JSON表示。
    """
    merged_results = []
    current_result = None

    for block in blocks:
        if current_result is None or (block.get('id') and block['id'] != current_result['id']):
            if current_result:
                merged_results.append(current_result)
            current_result = block.copy()
        else:
            for key, value in block.items():
                if isinstance(value, dict):
                    if key == 'function':
                        current_result[key] = merge_function_fields(current_result.get(key, {}), value)
                    else:
                        current_result[key] = merge_json_blocks([current_result.get(key, {}), value])
                elif isinstance(value, str):
                    if key in ['type'] and value:
                        if value != current_result.get(key, ''):
                            current_result[key] = value 
                    else:
                        current_result[key] = current_result.get(key, '') + value
                else:
                    current_result[key] = value

    if current_result:
        merged_results.append(current_result)

    return merged_results

def merge_json_blocks(blocks: List[Dict[str, Any]]) -> Dict[str, Any]:
    merged_result = {}

    for block in blocks:
        for key, value in block.items():
            if key not in merged_result:
                merged_result[key] = value
            else:
                if isinstance(value, dict) and isinstance(merged_result[key], dict):
                    if key == 'function':
                        merged_result[key] = merge_function_fields(merged_result[key], value)
                    else:
                        merged_result[key] = merge_json_blocks([merged_result[key], value])
                elif value != merged_result[key]:
                    merged_result[key] = value

    return merged_result

def merge_function_fields(original: Dict[str, Any], new: Dict[str, Any]) -> Dict[str, Any]:
    merged_function = original.copy()
    if 'name' in new:
        merged_function['name'] = original.get('name', '') + new['name']
    if 'arguments' in new:
        merged_function['arguments'] = original.get('arguments', '') + new['arguments']
    return merged_function

def merge_blocks_by_index(blocks: List[Dict[str, Any]]) -> Dict[int, List[Dict[str, Any]]]:
    index_groups = {}
    for block in blocks:
        index = block.get('index')
        if index is not None:
            if index not in index_groups:
                index_groups[index] = []
            index_groups[index].append(block)
    
    merged_results = {index: merge_tool_calls(group) for index, group in index_groups.items()}
    return merged_results

def count_tokens(text: str):
    """
    计算文本的 token 数量。
    """
    return len(get_token_ids(text))

def get_token_ids(text: str, token_encoding: str=None, allowed_special: str=None, disallowed_special: str=None) -> List[int]:
    """
    获取文本的 token ID 列表。
    """
    encoding_model = tiktoken.get_encoding(token_encoding or 'cl100k_base')
    return encoding_model.encode(
        text,
        allowed_special = allowed_special or set(),
        disallowed_special = disallowed_special or "all",
    )



import re
import copy
import os
import yaml
from typing import Iterator, List, Union, Optional

from ......config import get_env
from ......utils import extract_text
from .....document import Document
from .utils import parse_markdown, create_front_matter

class Markdown():
    def __init__(self, doc_str: str=None, md_file: str=None, source: str=None):
        self.documents = []
        self.doc_str = doc_str
        self.md_file = md_file
        self.source = source or md_file
        self.import_markdown(doc_str, md_file)

    def import_markdown(self, doc_str: str=None, md_file: str=None):
        """
        导入Markdown文档。
        """
        _doc_str = doc_str or self.doc_str
        _md_file = md_file or self.md_file
        if _md_file and os.path.isfile(_md_file) and os.path.exists(_md_file):
            with open(_md_file, 'r', encoding='utf-8') as file:
                _doc_str = file.read()

        if _doc_str:
            self.documents = parse_markdown(_doc_str, source=self.source)

        return self.documents
    
    @classmethod
    def to_text(cls, documents: List[Document], sep: str="", with_front_matter: bool=False):
        if not documents:
            return ''

        meta0 = documents[0].meta
        front_matter = ''
        if with_front_matter and 'type' in meta0 and meta0['type'] == 'front_matter':
            front_matter = create_front_matter(meta0)

        return front_matter + sep.join([d.text for d in documents])

    @property
    def text(self):
        return self.__class__.to_text(self.documents)

    def get_all(self, pattern: str=None):
        return [
            d
            for d in self.documents
            if re.search(pattern or '.*', d.text)
        ]

    def get_outline(self, pattern: str=None):
        return [
            d
            for d in self.documents
            if d.meta['type'] == "OUTLINE"
            and re.search(pattern or '.*', d.text)
        ]
    
    def fetch_outline_task(self, outline_doc: Document, prev_k: int=800, next_k: int=200):
        """
        提取提纲内容。

        返回结果为 (draft: str, task: str)，其中：
        - draft 为结合了前后文和扩写位置说明的草稿
        - task 为具体的要求
        """
        if outline_doc.meta['type'] != 'OUTLINE':
            raise  ValueError(f"Document's type Must be OUTLINE!")

        # 提取草稿
        docs = []

        docs.extend(self.get_prev_documents(outline_doc, k=prev_k))

        outline = Document(text="<<<YOUR_TEXT>>>\n\n", meta={"type": "paragraph"})
        docs.append(outline)

        docs.extend(self.get_next_documents(outline_doc, k=next_k))

        draft = self.__class__.to_text(docs)

        # 提取任务
        tag_start = get_env("ILLUFLY_OUTLINE_START")
        tag_end = get_env("ILLUFLY_OUTLINE_END")
        md = self.__class__.to_text([outline_doc])
        task = extract_text(md, (tag_start, tag_end))

        return (draft, task)

    def get_task_range(self, index_from: Union[str, Document], index_to: Union[str, Document]):
        """
        获得任务索引。
        """
        index_from = index_from.meta['id'] if isinstance(index_from, Document) else index_from
        index_to = index_to.meta['id'] if isinstance(index_to, Document) else index_to
        _from = None
        _to = None
        for i, doc in enumerate(self.documents):
            if doc.meta['id'] == index_from:
                _from = i
            if doc.meta['id'] == index_to:
                _to = i
            if _from != None and _to != None:
                return (_from, _to)
        return (_from, _to)

    def replace_documents(self, doc_from: Union[str, Document], doc_to: Union[str, Document], docs: Union[str, List[Document]]=None):
        """
        替换文档对象。
        
        如果被提取的文本与原有上文和下文重叠，就剔除重叠的部份。
        """
        to_insert = parse_markdown(docs) if isinstance(docs, str) else docs

        _from, _to = self.get_task_range(doc_from, doc_to)
        if _from == None:
            raise ValueError(f"{doc_from} NOT FOUND!")
        if _to == None:
            raise ValueError(f"{doc_to} NOT FOUND!")
        
        prev_docs = self.documents[:_from]
        next_docs = self.documents[_to + 1:]
        
        prev_heading = next((doc for doc in reversed(prev_docs) if 'type' in doc.meta and doc.meta['type'] == 'heading'), None)
        next_heading = next((doc for doc in next_docs if 'type' in doc.meta and doc.meta['type'] == 'heading'), None)

        from_index = None
        if prev_heading:
            for i, d in enumerate(to_insert):
                if d.text.strip() == prev_heading.text.strip():
                    from_index = i + 1
                    break

        to_index = None
        if next_heading:
            for i, d in enumerate(reversed(to_insert)):
                if d.text.strip() == next_heading.text.strip():
                    to_index = i
                    break

        to_insert_docs = to_insert[from_index:to_index]

        self.documents = prev_docs + to_insert_docs + next_docs
        return to_insert_docs

    def get_prev_documents(self, index_doc: Union[int, str, Document]=None, k: int=800):
        """
        获得向前关联文档。
        """
        if index_doc is None:
            return []

        task_id = index_doc.meta['id'] if isinstance(index_doc, Document) else index_doc

        md = ""
        found_task = False
        last_header_level = None
        docs = []

        for _doc in self.documents[::-1]:
            new_doc = copy.deepcopy(_doc)
            if not found_task:
                if new_doc.meta['id'] == task_id:
                    found_task = True
                continue

            # 在token数量可承受范围内优先前文
            # 并且，在获得上下文内容时，下文内容中不出现<OUTLINE>
            if new_doc.meta['type'] == 'OUTLINE':
                new_doc.text = '...\n'
            md = new_doc.text + md
            if len(md) <= k:
                docs.append(new_doc)
                continue

            # 补充所有祖先标题
            doc_is_header = new_doc.meta['type'] == "heading"
            if doc_is_header:
                doc_level = new_doc.meta['attrs']['level']
                if last_header_level == None or doc_level <= last_header_level:
                    docs.append(new_doc)
                    last_header_level = doc_level

        return docs[::-1]

    def get_next_documents(self, index_doc: Union[str, Document]=None, k:int =200):
        """
        获得向后关联文档。
        
        推理能力较弱的模型，不可在上下文中出现 <OUTLINE></OUTLINE> 包含的扩写内容。
        否则模型将无视提示语中无需的扩写指令，对其展开扩写。
        """
        if index_doc is None:
            return []

        task_id = index_doc.meta['id'] if isinstance(index_doc, Document) else index_doc

        md = ""
        found_task = False
        docs = []

        for _doc in self.documents:
            new_doc = copy.deepcopy(_doc)
            if not found_task:
                if new_doc.meta['id'] == task_id:
                    found_task = True
                continue

            # 获得上下文内容时，下文内容中不出现<OUTLINE>
            if new_doc.meta['type'] == 'OUTLINE':
                new_doc.text = '...\n'
            md = new_doc.text + md
            if len(md) <= k:
                docs.append(new_doc)
            else:
                break

        return docs


import os
import json
import copy

from typing import Union, List

from .base import BaseHistory

class InMemoryHistory(BaseHistory):
    """基于内存的记忆管理"""

    def __init__(self, memory: dict = {}, **kwargs):
        super().__init__(**kwargs)
        self.memory = memory or {}

    def last_thread_id_count(self):
        all_thread_ids = self.list_threads()
        if all_thread_ids:
            ids = all_thread_ids[-1].split("-")
            return int(ids[-1]) + 1
        else:
            return 0

    # 列举所有记忆线
    def list_threads(self):
        return sorted(self.memory.keys())

    def save_memory(self, thread_id: str, memory: List[dict]):
        self.memory[thread_id] = copy.deepcopy(memory)

    def load_memory(self, thread_id: Union[str, int] = None):
        """
        加载记忆。

        如果 thread_id 是字符串，则直接加载指定线程的记忆；
        如果 thread_id 是整数，则将其当作索引，例如 thread_id=-1 表示加载最近一轮对话的记忆。
        """
        _thread_id = thread_id
        if isinstance(thread_id, str):
            return _thread_id, self.memory.get(thread_id, [])
        elif isinstance(thread_id, int):
            all_threads = self.list_threads()
            if all_threads:
                _thread_id = all_threads[thread_id]
                return _thread_id, self.memory.get(_thread_id, [])

        return _thread_id, []



from typing import List, Union, Dict, Any
from ......io import EventBlock, NewLineBlock
from ....message import Messages
from ...base import BaseAgent

class BaseToolCalling:
    def __init__(
        self,
        tools_to_exec: List[BaseAgent] = None,
        steps: List[Any] = None
    ):
        self.tools_to_exec = tools_to_exec
        self.steps = steps or []

    def reset(self, tools_to_exec: List[BaseAgent]=None):
        self.tools_to_exec = tools_to_exec or self.tools_to_exec
        self.steps.clear()

    @property
    def completed_work(self):
        """
        返回所有步骤的完成情况。
        """
        return "\n".join([str(step) for step in self.steps])

    def extract_tools_call(self, text: str) -> List[Dict[str, Any]]:
        """
        解析工具回调。

        text 必须为包含工具回调描述的字符串。
        """
        return []
    def handle(self, final_tools_call):
        """
        处理工具回调。

        final_tools_call 是一组自定义的工具解析结果，这可能是工具回调描述、计划描述的列表，也可以是DAG描述等形式。
        """
        return []

    async def async_handle(self, final_tools_call):
        return []

    def execute_tool(self, tool):
        """
        执行工具回调。

        tool 必须具有如下 Python 字典格式：
        {
            "function":
                "name": function_name
                "arguments": arguments_json
        }
        其中 function_name 必须为 self.tools_to_exec 中的某个 name 属性，
        arguments 则必须为一个可转换为 Python 字典的 JSON 字符串。
        """
        if not self.tools_to_exec:
            yield EventBlock("warn", f"tools_to_exec is empty")
            return

        for struct_tool in self.tools_to_exec:
            if tool.get('function', {}).get('name') == struct_tool.name:
                yield EventBlock("agent", struct_tool.name)
                tool_args = struct_tool.parse_arguments(tool['function']['arguments'])
                tool_resp = ""

                if isinstance(tool_args, dict):
                    tool_func_result = struct_tool.call(**tool_args)
                elif isinstance(tool_args, list):
                    tool_func_result = struct_tool.call(*tool_args)
                else:
                    yield EventBlock("action_parse_failed", tool_args)
                for x in tool_func_result:
                    if isinstance(x, EventBlock):
                        if x.block_type == "final_tool_resp":
                            tool_resp = x.text
                        elif x.block_type in ["chunk", "text"]:
                            tool_resp += x.text
                        yield x
                    else:
                        tool_resp += x
                        yield EventBlock("tool_resp_chunk", x)
                yield NewLineBlock()
                if tool_resp:
                    yield EventBlock("final_tool_resp", tool_resp)
                return

        yield EventBlock("warn", f"tool {tool} not found")
            

    async def async_execute_tool(self, tool):
        for struct_tool in self.tools_to_exec:
            if tool.get('function', {}).get('name') == struct_tool.name:
                yield EventBlock("agent", struct_tool.name)
                tool_args = struct_tool.parse_arguments(tool['function']['arguments'])
                tool_resp = ""

                if isinstance(tool_args, dict):
                    tool_func_result = struct_tool.async_call(**tool_args)
                elif isinstance(tool_args, list):
                    tool_func_result = struct_tool.async_call(*tool_args)
                else:
                    yield EventBlock("action_parse_failed", tool_args)
                async for x in tool_func_result:
                    if isinstance(x, EventBlock):
                        if x.block_type == "final_tool_resp":
                            tool_resp = x.text
                        elif x.block_type in ["chunk", "text"]:
                            tool_resp += x.text
                        yield x
                    else:
                        tool_resp += x
                        yield EventBlock("tool_resp_chunk", x)

                yield NewLineBlock()
                if tool_resp:
                    yield EventBlock("final_tool_resp", tool_resp)



import json
import copy
import os
from typing import Union, Dict, Any, List, Tuple

from ..runnable import Runnable

class Message:
    def __init__(self, role: str, content: Union[str, Runnable], **kwargs):
        self.role = role
        self.content = content
        self.kwargs = kwargs

    def __str__(self):
        return f"{self.role}: {self.content.selected.format() if isinstance(self.content, Runnable) else self.content}"

    def __repr__(self):
        return f"Message(role={self.role}, content={self.content})"
    
    def to_dict(self, binding: Dict[str, Any]=None, style: str=None):
        def get_url(item, key, alt_key):
            return item.get(key, item[alt_key]['url'] if alt_key in item and 'url' in item[alt_key] else None)
        
        def get_unique_format(style, binding):
            if isinstance(self.content, list):
                contents = []
                for item in self.content:
                    content = {}
                    if isinstance(item, str):
                        content["text"] = item
                    elif isinstance(item, dict):
                        if item.get("audio") or "audio_url" in item:
                            content["audio"] = item.get("audio", get_url(item, "audio", "audio_url"))
                        if item.get("video") or "video_url" in item:
                            content["video"] = item.get("video", get_url(item, "video", "video_url"))
                        if item.get("image") or "image_url" in item:
                            content["image"] = item.get("image", get_url(item, "image", "image_url"))
                        if item.get("text") is not None:
                            content["text"] = item.get("text")
                    else:
                        raise ValueError("Resource must be a string or a dict")
                    contents.append(content)
                return contents
            elif isinstance(self.content, str):
                return [{"text": self.content}]
            elif isinstance(self.content, Runnable):
                return [{"text": self.content.selected.format(binding)}]
        
        if style == "openai_vl":
            content = [
                {
                    "type": k if k == "text" else k + "_url",
                    (k if k == "text" else k + "_url"): v if k == "text" else {"url": v}
                }
                for item in get_unique_format(style, binding)
                for k, v in item.items()
            ]
        elif style == "qwen_vl":
            content = get_unique_format(style, binding)
        else:
            unique_format = get_unique_format(style, binding)
            if unique_format:
                content = "\n".join([c['text'] for c in unique_format if 'text' in c])
            else:
                content = ""

        return {
            "role": self.role,
            "content": content,
            **self.kwargs
        }

    def to_text(self, binding: Dict[str, Any]=None):
        if isinstance(self.content, Runnable):
            return self.content.selected.format(binding)
        return self.content

    def to_json(self, binding: Dict[str, Any]=None):
        if isinstance(self.content, Runnable):
            content = self.content.selected.format(binding)
        else:
            content = self.content
        return json.dumps({
            "role": self.role,
            "content": self.content,
            **self.kwargs
        })

class Messages:
    """
    Messages 类被设计为「隐藏类」，用来简化消息列表的管理，但并不作为交换类型出现在流程中。
    这主要是为了回避增加不必要的新概念。

    构造函数中的 messages 参数将作为原始的「只读对象」，被保存到 self.raw_messages 中。
    无论你提供的列表元素是字符串、模板、元组还是字典，都会在实例化完成后，生成 Message 对象列表，并被写入到 self.messages 中。
    但类似 {"role": xx, "content": yy} 这种消息列表，是通过 self.to_list 方法「动态提取」的，这主要是为了兼容 PromptTemplate 对象的模板变量。
    """

    def __init__(
        self,
        messages: Any=None,
        style: str=None,
        binding: Dict[str, Any]=None
    ):
        self.raw_messages = messages or []
        self.style = style
        self.binding = binding or {}
        if not isinstance(self.raw_messages, list):
            self.raw_messages = [self.raw_messages]

        self.messages = []
        for i, msg in enumerate(self.raw_messages):
            self.messages.append(self._convert_to_message(msg, i))

    def _convert_to_message(self, msg: Any, index: int) -> Message:
        message = None
        if isinstance(msg, Message):
            message = msg
        elif isinstance(msg, Runnable):
            message = Message(role='system' if index == 0 else self._determine_role(index), content=msg)
        elif isinstance(msg, str):
            message = Message(role='user' if index == 0 else self._determine_role(index), content=msg)
        elif isinstance(msg, list):
            role = self._determine_role(index)
            message = Message(role=role, content=msg)
        elif isinstance(msg, dict):
            if msg.get('role') == 'ai':
                msg['role'] = 'assistant'
            message = Message(**msg) # 支持字典构造中其他键值，如工具回调等
        elif isinstance(msg, tuple):
            if len(msg) == 2 and msg[0] in ["ai", "assistant", "user"]:
                role, content = msg
                if role == 'ai':
                    role = 'assistant'
                message = Message(role=role, content=content)
            else:
                # 多模态格式
                role = self._determine_role(index)
                msgs = []
                for item in msg:
                    if isinstance(item, str):
                        if item in ["ai", "assistant"]:
                            role = "assistant"
                        elif item in ["user"]:
                            role = "user"
                        else:
                            ext = os.path.splitext(item)[1][1:]
                            if ext.lower() in ["png", "jpg", "gif", "jpeg"]:
                                msg_type = "image"
                            elif ext.lower() in ["mp4", "webm"]:
                                msg_type = "video"
                            else:
                                msg_type = "text"
                            msgs.append({
                                msg_type: item
                            })
                    # elif isinstance(item, Runnable):
                    #     msgs.append({"text": item})
                    else:
                        raise ValueError("Unsupported message type in tuple", msg)
                message = Message(role=role, content=msgs)
        else:
            raise ValueError("Unsupported message type", msg)
        
        return message

    def _determine_role(self, index: int) -> str:
        if index == 0:
            return 'system'
        elif self.messages[0].role == 'system' and index == 1:
            return 'user'
        else:
            last_role = self.messages[-1].role if self.messages else 'user'
            return 'assistant' if last_role == 'user' else 'user'

    def clone(self):
        return Messages(copy.deepcopy(self.raw_messages), style=self.style, binding=self.binding)

    def __str__(self):
        return "\n".join([str(message) for message in self.messages])

    def __repr__(self):
        return f"<Messages({self.length} items)>"
    
    def __getitem__(self, index: int):
        return self.messages[index]
    
    def __iter__(self):
        return iter(self.messages)
    
    def __len__(self):
        return self.length

    def __add__(self, other):
        if not isinstance(other, Messages):
            raise TypeError("Operands must be of type Messages")
        combined_messages = self.messages + other.messages
        return Messages(combined_messages, style=self.style, binding=self.binding)
    
    def to_list(self, binding: Dict[str, Any]=None, style: str=None):
        return [msg.to_dict({**self.binding, **(binding or {})}, (style or self.style)) for msg in self.messages]
    
    def to_json(self, binding: Dict[str, Any]=None, style: str=None):
        return [msg.to_json({**self.binding, **(binding or {})}, (style or self.style)) for msg in self.messages]
    
    @property
    def length(self):
        return len(self.messages)

    def append(self, message: Union[Message, str, Runnable, dict, Tuple[str, Union[str, Runnable]]]):
        self.messages.append(self._convert_to_message(message, len(self.messages)))

    def extend(self, messages: Union[Message, str, Runnable, Tuple[str, Union[str, Runnable]], List[Union[Message, str, Runnable, dict, Tuple[str, Union[str, Runnable]]]]]):
        for msg in messages:
            self.append(msg)

    @property
    def last_role(self):
        if self.messages:
            return self.messages[-1].role
        else:
            return None
    
    @property
    def last_content(self):
        if self.messages:
            return self.messages[-1].content
        else:
            return None

    @property
    def all_templates(self):
        """
        从消息列表中提取所有模板对象
        """
        if self.messages:
            return [msg.content for msg in self.messages if isinstance(msg.content, Runnable)]
        else:
            return []

    def has_role(self, role: str):
        return any(msg.role == role for msg in self.messages)



import os
import json
import time
import random

from typing import Union, List, Dict, Any, Callable
from .....utils import filter_kwargs, raise_invalid_params, get_env
from ....history import thread_id_gen, BaseHistory, LocalFileHistory, InMemoryHistory
from ...message import Messages, Message
from ...prompt_template import PromptTemplate
from ...binding_manager import BindingManager

class MemoryManager(BindingManager):
    @classmethod
    def allowed_params(cls):
        """
        返回当前可用的参数列表。
        """
        return {
            "style": "消息样式",
            "memory": "记忆列表",
            "remember_rounds": "记忆轮数",
            "history": "记忆持久化管理",
            **BindingManager.allowed_params(),
        }

    def __init__(
        self,
        style: str=None,
        memory: Union[List[Union[str, "PromptTemplate", Dict[str, Any]]], Messages]=None,
        remember_rounds: int=None,
        history: BaseHistory=None,
        **kwargs
    ):
        raise_invalid_params(kwargs, self.allowed_params())
        super().__init__(**filter_kwargs(kwargs, BindingManager.allowed_params()))

        self.style = style
        self.memory = []
        self.remember_rounds = remember_rounds if remember_rounds is not None else 10

        self.history = history or InMemoryHistory()
        self.history.reset_init(self.__class__.__name__, self.name)
        # 加载最近一轮对话的记忆
        self.history.load_memory(self.history.last_thread_id)

        self.init_memory = []
        self.reset_init_memory(memory)
        self._thread_id = None
        self._current_thread_size = len(self.memory)
        self._chat_memory = []

    @property
    def thread_ids(self):
        return self.history.list_threads()

    @property
    def current_thread_size(self):
        return self._current_thread_size

    @property
    def thread_id(self):
        return self._thread_id

    def create_thread_id(self):
        last_count = self.history.last_thread_id_count()
        self._current_thread_size = 0
        self._thread_id = thread_id_gen.create_id(last_count)
        return self._thread_id

    @property
    def chat_memory(self):
        return self._chat_memory

    def create_new_thread(self):
        """
        开启新一轮对话
        """
        self.memory.clear()
        if self._thread_id is None or self.current_thread_size > 0:
            self._thread_id = next(self.create_thread_id())
        return self._thread_id

    def reset_init_memory(self, messages: Union[str, List[dict]]):
        self.init_memory = Messages(messages, style=self.style)
        for template in self.init_memory.all_templates:
            self.bind_consumer(template)

    def get_bound_vars(self, new_messages: Messages, new_chat: bool=False):
        """
        获取所有被 PromptTemplates 绑定的变量。

        这主要用于判断 provider_dict 中的键值是否被消息列表中的模板绑定使用。
        典型的场景包括：判断 input 和 tools_desc 是否被使用，这将会影响对话过程中组织对话或提供 tools 参数给大模型。
        """
        _new_messages = self.init_memory
        if new_chat and new_messages.has_role("system"):
            _new_messages = new_messages
        else:
            _new_messages += new_messages

        bound_vars = set()
        for template in _new_messages.all_templates:
            _template = template.selected
            _template_vars = _template.template_vars
            bound_vars.update(_template_vars)
            for provider_key, consumer_key in _template.lazy_binding_map.items():
                if provider_key in _template_vars:
                    bound_vars.remove(provider_key)
                    if consumer_key in self.provider_dict:
                        bound_vars.add(consumer_key)
        return bound_vars

    # 保存记忆
    def save_memory(self):
        self.history.save_memory(self.thread_id, self.memory)
        self._current_thread_size = len(self.memory)

    # 加载记忆
    def load_memory(self, thread_id: Union[str, int]=None):
        """
        加载指定记忆。
        - 使用整数索引: -1 加载最近一轮对话的记忆; 0 加载首轮记忆
        - 使用字符串: 加载指定记忆 thread_id
        """
        self._thread_id, self.memory = self.history.load_memory(thread_id)
        self._current_thread_size = len(self.memory)

    def build_chat_memory(self, prompt: Union[str, List[dict]], new_chat: bool=False, remember_rounds: int = None):
        """
        获取对话所需的短时记忆。

        主要用于构建对话时的历史记忆，并将其作为提示语中上下文的一部分。
        具体包括：
        - 根据问题场景，确定是否建立新一轮对话
        - 根据问题，构造提问对话，并补充模板变量
        - 根据记忆，捕捉已有对话记录
        - 根据所提供的知识背景，构造知识背景对话
        - 将以上对话合并，构造完整的对话场景短时记忆
        """

        # 构建 prompt 标准形式
        new_prompt = prompt if isinstance(prompt, Messages) else Messages(prompt, style=self.style)

        # 确定是否新一轮对话
        if new_chat or new_prompt.has_role("system") or not self.memory:
            new_chat = True
            self.create_new_thread()

        new_messages = self.build_new_messages(new_prompt, new_chat)
        history_memory = self.get_history_memory(new_chat, remember_rounds)

        if new_messages.has_role("system"):
            new_messages_list = Messages((new_messages[:1] + history_memory.messages + new_messages[1:]), style=self.style).to_list()
            self.memory.extend(new_messages_list)
            self._chat_memory = new_messages_list
            self._current_thread_size = len(self.memory)
        else:
            self.memory.extend(new_messages.to_list())
            self._chat_memory = (history_memory + new_messages).to_list()
            self._current_thread_size = len(self.memory)
        return self._chat_memory

    def get_history_memory(self, new_chat: bool=False, remember_rounds: int = None):
        """
        获取历史记忆。
        """
        _k = self.remember_rounds if remember_rounds is None else remember_rounds
        final_k = 2 * _k if _k >= 1 else 1
        if len(self.memory) > 0 and self.memory[0]['role'] == 'system':
            new_messages = self.memory[:1]
            new_messages += self.memory[1:][-final_k:]
        else:
            new_messages = self.memory[-final_k:]
        return Messages(new_messages, style=self.style)

    def build_new_messages(self, prompt: Union[str, List[dict]], new_chat: bool=False, **kwargs):
        """
        构建提问消息列表。
        """
        # 构建 prompt 标准形式
        if isinstance(prompt, str):
            prompt = [{"role": "user", "content": prompt}]
        new_messages = prompt if isinstance(prompt, Messages) else Messages(prompt, style=self.style)

        # 无论新消息列表中是否包含 system 角色，都需要绑定模板变量，但应当是��态绑定
        templates = new_messages.all_templates
        for template in templates:
            self.bind_consumer(template, dynamic=True)

        if new_chat:
            if not new_messages.has_role("system"):
                # 只要没有提供 system 角色，就启用 init_memory 模板
                # 合并 new_messages 和 init_memory
                new_messages = self.init_memory + new_messages

            # 如果在合并后的 new_messages 中，task 变量被模板使用，
            # 则将尾部消息列表中的 user 角色消息取出，并赋值给 task 变量用于绑定映射
            bound_vars = self.get_bound_vars(new_messages, new_chat=new_chat)
            if 'task' in bound_vars and new_messages[-1].role == 'user':
                new_messages.messages.pop(-1)
            
            # 如果新消息列表的尾部是 system 消息，则需要补充一个 user 角色消息
            # 否则，缺少用户消息会让大模型拒绝回答任何问题
            if new_messages[-1].role == 'system':
                new_messages.append({"role": "user", "content": "请开始"})

        return new_messages

    def remember_response(self, response: Union[str, List[dict]]):
        """
        将回答添加到记忆中。
        """
        if response:
            if isinstance(response, str):
                new_memory = Messages([("assistant", response)], style=self.style).to_list()
            else:
                new_memory = response

            self.memory.extend(new_memory)
            self._current_thread_size = len(self.memory)
            return new_memory
        else:
            return []



import json

from typing import Union, List, Dict, Any
from ..base import BaseAgent
from .tools_calling import BaseToolCalling, ToolCall, Plans, SubTask

class ToolsManager:
    """
    ToolsManager 基类，用于管理工具相关的功能。

    在实际使用中, tools 参数可能被大模型作为参数引用，也可能被提示语模板引用。
    如果 tools_desc 被提示语模板绑定，则 tools 就从传递给大模型的参数中剔除。

    tools_handlers 是工具处理器的列表，每个处理器可以处理一种工具调用行为。

    tools_behavior 是工具处理器的处理行为列表，有三种值：
    - parse: 仅解析工具回调, 不执行
    - parse-execute: 解析工具回调后执行, 然后停止
    - parse-execute-continue: 解析工具回调, 执行工具, 然后继续调用模型
    """
    @classmethod
    def allowed_params(cls):
        """
        返回当前可用的参数列表。
        """
        return {
            "tools": "工具列表",
            "tools_handlers": "工具处理器列表",
            "tools_behavior": "工具处理行为, 包括 parse-execute, parse-execute-continue, parse-continue-execute 三种行为",
        }

    def __init__(
        self,
        tools=None,
        tools_handlers: List[BaseToolCalling]=None,
        tools_behavior: str=None
    ):
        self.tools = [
            t if isinstance(t, BaseAgent) else BaseAgent(t)
            for t in (tools or [])
        ]
        self.tools_handlers = tools_handlers or [ToolCall(), Plans(), SubTask()]
        for h in self.tools_handlers:
            h.reset(self.tools)

        self.tools_behavior = tools_behavior or "parse-execute-continue"

    def get_tools(self, tools: List["BaseAgent"]=None):
        _tools = [
            t if isinstance(t, BaseAgent) else BaseAgent(func=t)
            for t in (tools or [])
        ]
        if _tools and (
            not isinstance(_tools, list) or
            not all(isinstance(tool, BaseAgent) for tool in _tools)
        ):
            raise ValueError("tools 必须是 BaseAgent 列表")
        return self.tools + _tools

    def get_tools_name(self, tools: List[BaseAgent]=None):
        return ",".join([a.name for a in self.get_tools(tools)])

    def get_tools_desc(self, tools: List[BaseAgent]=None):
        return [t.tool_desc for t in self.get_tools(tools)]



import os
from concurrent.futures import ThreadPoolExecutor

class ExecutorManager:
    """
    对于使用多线程实现的外部调用，可以在环境变量中配置默认的线程池数量。
    例如：
    DEFAULT_MAX_WORKERS_CHAT_OPENAI=10
    """
    executors = {}

    @classmethod
    def allowed_params(cls):
        """
        返回当前可用的参数列表。
        """
        return {
            "threads_group": "如果由 illufly 管理线程池实现并发或异步，则可以指定线程组名称，默认为 DEFAULT",
        }

    def __init__(self, threads_group: str = None, **kwargs):
        self.threads_group = threads_group or "DEFAULT"
        if self.threads_group not in self.executors:
            max_workers = int(os.getenv(f"DEFAULT_MAX_WORKERS_{self.threads_group.upper()}", 5))
            self.executors[self.threads_group] = ThreadPoolExecutor(max_workers=max_workers)
        self.executor = self.executors[self.threads_group]

    @classmethod
    def monitor_executors(cls):
        info = {}
        for group, executor in cls.executors.items():
            active_threads = len(executor._threads)
            max_workers = executor._max_workers
            waiting_threads = executor._work_queue.qsize()
            info[group] = {
                "max_workers": max_workers,
                "used_workers": active_threads,
                "waiting_threads": waiting_threads
            }
        return info

    @classmethod
    def shutdown_executors(cls):
        for executor in cls.executors.values():
            executor.shutdown(wait=True)


from typing import List, Dict

from .....utils import extract_segments, raise_invalid_params
from .....io import EventBlock
from ....dataset import Dataset
from ...prompt_template import PromptTemplate
from ..base import BaseAgent
from ..chat import ChatAgent

import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import platform
import matplotlib.font_manager as fm

class PythonAgent(BaseAgent):
    """
    处理 Python 相关任务的基础 Agent。
    """
    @classmethod
    def allowed_params(cls):
        return {
            "agent": "生成代码的 ChatAgent 实例",
            "datasets": "数据集清单, 类型为 Dataset 实例，或者提供 `{name: Dataset}` 的键值对来创建 Dataset 实例",
            "exec_code": "是否执行生成的代码，默认为 True",
            "font_path": "字体路径，用于绘制图表时使用",
            **BaseAgent.allowed_params(),
        }

    def __init__(
        self,
        agent: ChatAgent,
        datasets: List[Dataset]=None,
        exec_code: bool=True,
        font_path: str=None,
        **kwargs
    ):
        raise_invalid_params(kwargs, self.allowed_params())
        super().__init__(**kwargs)

        self.datasets = self._initialize_datasets(datasets)
        self.exec_code = exec_code

        if not isinstance(agent, ChatAgent):
            raise ValueError("agent 必须是 ChatAgent 实例")

        self.agent = agent
        self.agent.start_marker = "```python"
        self.agent.end_marker = "```"

        self._last_code = None

        self.reset_datasets()

    def _initialize_datasets(self, datasets):
        if datasets:
            if isinstance(datasets, dict):
                if not all(isinstance(ds, Dataset) for ds in datasets.values()):
                    raise ValueError("datasets 每个值都必须是 `Dataset` 类型")
                return datasets
            elif isinstance(datasets, list):
                if not all(isinstance(ds, Dataset) for ds in datasets):
                    raise ValueError("datasets 每个值都必须是 `Dataset` 类型")
                return {ds.name: ds for ds in datasets}
            else:
                raise ValueError("datasets 必须是 `Dataset` 实例的列表")
        return {}

    def reset_datasets(self):
        """
        重置数据集描述
        """
        self.agent.reset_init_memory(self.prompt_template)
        self.description = self.datasets_description

    def add_dataset(self, df: pd.DataFrame, name: str, desc: str=None):
        """
        添加数据集
        """
        self.datasets[name] = Dataset(df, name, desc or name)

    @property
    def datasets_description(self):
        """
        数据集描述
        """
        names = ', '.join(self.datasets.keys())
        descs = '\n'.join([f"- {name}: {ds.desc}" for name, ds in self.datasets.items()])
        return f"回答关于[{names}]等数据集的相关问题。\n这些数据集具体包括：{descs}"

    @property
    def datasets_summary(self):
        """
        数据集摘要
        """
        return '\n'.join([ds.summary for ds in self.datasets.values()])

    @property
    def datasets_description(self):
        """
        数据集描述
        """
        names = ', '.join(self.datasets.keys())
        descs = '\n'.join([f"- {name}: {ds.desc}" for name, ds in self.datasets.items()])
        return f"回答关于[{names}]等数据集的相关问题。\n这些数据集具体包括：{descs}"

    @property
    def datasets_summary(self):
        """
        数据集摘要
        """
        return '\n'.join([ds.summary for ds in self.datasets.values()])

    @property
    def prompt_template(self):
        """
        提示语模板
        """
        return PromptTemplate(
            self.template_id,
            binding_map={
                "registered_global": lambda: list(self.registered_global.keys()),
                "safe_builtins": lambda: list(self.safe_builtins.keys()),
                "dataset_names": lambda: ', '.join(self.datasets.keys()),
                "dataset_description": lambda: self.datasets_description,
                "dataset_summary": lambda: self.datasets_summary,
                "safe_header_code": lambda: self.safe_header_code,
            })

    @property
    def last_code(self):
        """
        上一次生成的代码
        """
        return self._last_code

    def execute_code(self, code: str):
        """
        执行代码，并返回执行结果。
        """
        # 创建一个新的无污染空间
        exec_namespace = self.registered_global.copy()
        try:
            exec(code, exec_namespace)
        except Exception as e:
            return f"执行代码时发生错误: {e}"

        return exec_namespace.get('last_output', "生成的代码已经执行，但返回了空结果。")

    @property
    def registered_global(self):
        """
        注册全局变量
        """
        return {
            "__builtins__": self.safe_builtins,
            "last_output": None,
            "datasets": self.datasets,  # 数据集清单
            "add_dataset": self.add_dataset,  # 添加数据集
        }

    @property
    def safe_header_code(self):
        """
        安全模块
        """
        return "\n".join([
            "import pandas as pd",
            "import numpy as np",
            "import seaborn as sns",
            "import matplotlib.pyplot as plt",
            "import matplotlib.font_manager as fm",
            *self.get_set_chinese_font_code(),
        ])

    def get_set_chinese_font_code(self):
        """
        根据操作系统返回设置中文字体的代码。
        """
        os_name = platform.system()
        if os_name == 'Darwin':  # macOS
            font_path = '/System/Library/Fonts/STHeiti Light.ttc'
        elif os_name == 'Windows':  # Windows
            font_path = 'C:/Windows/Fonts/simsun.ttc'
        elif os_name == 'Linux':  # Linux
            font_path = '/usr/share/fonts/truetype/arphic/ukai.ttc'
        else:
            raise RuntimeError("Unsupported operating system for setting Chinese font.")

        return [
            f"font_prop = fm.FontProperties(fname='{font_path}')",
            "plt.rcParams['font.family'] = font_prop.get_name()",
        ]

    @property
    def safe_builtins(self):
        """
        安全内置函数
        """
        return {
            "__import__": __import__,
            "abs": abs,
            "all": all,
            "any": any,
            "bin": bin,
            "bool": bool,
            "chr": chr,
            "str": str,
            "complex": complex,
            "divmod": divmod,
            "enumerate": enumerate,
            "filter": filter,
            "hex": hex,
            "isinstance": isinstance,
            "issubclass": issubclass,
            "iter": iter,
            "len": len,
            "list": list,
            "map": map,
            "max": max,
            "min": min,
            "next": next,
            "oct": oct,
            "ord": ord,
            "pow": pow,
            "range": range,
            "reversed": reversed,
            "round": round,
            "sorted": sorted,
            "sum": sum,
            "tuple": tuple,
            "dict": dict,
            "set": set,
            "frozenset": frozenset,
            "zip": zip,
            "hasattr": hasattr,
            "print": print,
            "format": format,
            "frozenset": frozenset,
            "slice": slice,
            "type": type,
            "zip": zip,
            "classmethod": classmethod,
            "staticmethod": staticmethod,
            "property": property,
            "super": super,
            "issubclass": issubclass,
            "isinstance": isinstance,
            "iter": iter,
            "next": next,
        }

    def call(self, question: str, **kwargs):
        self._last_output = ""
        self._last_code = ""

        new_chat = kwargs.pop("new_chat", True)
        self.agent.reset_init_memory(self.prompt_template)
        yield from self.agent.call(question, **kwargs, new_chat=new_chat)

        # 禁止代码中执行引入其他库的操作
        safety_code = '\n'.join([
            line
            for line
            in self.agent.last_output.split('\n')
            if not line.strip().startswith('import')
        ])
        self._last_code = f"{self.safe_header_code}\n{safety_code}\nlast_output = main()\n"

        try:
            if self.exec_code:
                if self.last_code:
                    result = self.execute_code(self.last_code)
                    if isinstance(result, pd.DataFrame):
                        self._last_output = result.to_markdown()
                    else:
                        self._last_output = str(result)
                    yield EventBlock("text", self._last_output)
                else:
                    yield EventBlock("warn", "没有正确生成python代码, 执行失败。")
            else:
                self._last_output = self.agent.last_output
        except Exception as e:
            self._last_output = str(e)
            yield EventBlock("text", self._last_output)



from typing import List, Union
from ..base import Runnable
from ....core.document import Document
from ....utils import hash_text, clean_filename, raise_invalid_params
from ....config import get_env
from ....io import EventBlock

import os
import pickle

class BaseEmbeddings(Runnable):
    """
    句子嵌入模型。

    使用向量模型，将文本转换为向量，以便入库或查询。
    Document(text, meta={'source': str}) -> Document(text, meta={'embeddings': Vectors})

    例如：
    ```
    from illufly.types import Document
    from illufly.embeddings import TextEmbeddings # 通义千问文本向量模型

    embeddings = TextEmbeddings()
    docs = [Document("这是一个测试文本", meta={"source": "test"})]
    embeddings(docs)
    print(embeddings.last_output[0].meta['embeddings'])
    ```
    """
    @classmethod
    def allowed_params(cls):
        return {
            "model": "文本嵌入模型的名称",
            "base_url": "BASE_URL",
            "api_key": "API_KEY",
            "dim": "编码时使用的向量维度",
            "max_lines": "每次编码时处理的最大行数",
            **Runnable.allowed_params()
        }

    def __init__(self, model: str=None, base_url: str=None, api_key: str=None, dim: int=None, max_lines: int=None, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        super().__init__(**kwargs)
        self.dim = dim
        self.model = model
        self.base_url = base_url
        self.api_key = api_key
        self.max_lines = max_lines or 5
        self.clear_output()

    def clear_output(self):
        self._last_output = []

    def query(self, text: str, *args, **kwargs) -> List[float]:
        """将文本转换为向量，以便查询"""

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """将文本转换为向量，以便入库"""

    def _get_embeddings_folder(self):
        return os.path.join(get_env("ILLUFLY_CACHE_EMBEDDINGS"), self.__class__.__name__, self.model)

    def call(
        self,
        docs: Union[str, List[str], List[Document]],
        **kwargs
    ) -> List[Document]:
        """
        将文本字符串或 Document 类型，转换为带有文本向量的 Document 列表。

        如果提供参数是字符串，则表示查询模式，自动填写 source 为 '__query__'。
        (这可能在使用某些模型时有必要，例如通义千问的 embedding-v2 以下版本)
        """
        self._last_output = []
        if isinstance(docs, str):
            docs = [Document(docs, meta={'source': '__query__'})]
        elif isinstance(docs, Document):
            docs.meta['source'] = '__query__'

        if not isinstance(docs, list):
            raise ValueError("docs 必须是字符串或 Document 类型列表，但实际为: {type(docs)}")

        vector_folder = self._get_embeddings_folder()

        for index, d in enumerate(docs):
            if isinstance(d, str):
                docs[index] = Document(d)
            elif not isinstance(d, Document):
                raise ValueError(f"文档类型错误: {type(d)}")

        yield from self._process_batch(docs, vector_folder)

        self._last_output = docs
        return docs

    def _process_batch(self, docs, vector_folder):
        batch_texts = []
        batch_docs = []

        for i in range(0, len(docs), self.max_lines):
            batch = docs[i:i + self.max_lines]
            batch_texts = [d.text for d in batch]
            batch_docs = batch

            # 检查哪些文件已经存在
            existing_files = self._check_existing_files(batch_docs, batch_texts, vector_folder)
            batch_texts = [text for text, exists in zip(batch_texts, existing_files) if not exists]
            batch_docs = [doc for doc, exists in zip(batch_docs, existing_files) if not exists]

            if batch_texts:
                yield EventBlock("info", f"文本向量转换 {sum(len(d.text) for d in batch_docs)} 字 / {len(batch_docs)} 个文件")
                vectors = self.embed_documents(batch_texts)
                yield from self._save_vectors_to_cache(batch_docs, batch_texts, vectors, vector_folder)

    def _check_existing_files(self, docs, batch_texts, vector_folder):
        existing_files = []
        for index, text in enumerate(batch_texts):
            vector_path = hash_text(text) + ".emb"
            source = clean_filename(docs[index].meta['source'])
            cache_path = os.path.join(vector_folder, source or "no_source", vector_path)
            if os.path.exists(cache_path):
                with open(cache_path, 'rb') as f:
                    docs[index].meta['embeddings'] = pickle.load(f)
                existing_files.append(True)
            else:
                existing_files.append(False)
        return existing_files

    def _save_vectors_to_cache(self, docs, batch_texts, vectors, vector_folder):
        for index, text in enumerate(batch_texts):
            vector_path = hash_text(text) + ".emb"
            source = clean_filename(docs[index].meta['source'])
            cache_path = os.path.join(vector_folder, source or "no_source", vector_path)
            os.makedirs(os.path.dirname(cache_path), exist_ok=True)
            with open(cache_path, 'wb') as f:
                pickle.dump(vectors[index], f)
                docs[index].meta['embeddings'] = vectors[index]
                yield EventBlock('info', f'wrote embedding cache {cache_path} {text[0:50]}{"..." if len(text) > 50 else ""}')



from typing import Union, List, Optional, Dict, Any, Callable

from .....io import EventBlock, NewLineBlock
from .....utils import count_tokens
from .base import ChatAgent

import os
import json

class ChatPool(ChatAgent):
    def __init__(self, *agents, **kwargs):
        self.agents_pool = {a.name: a for a in list(agents)}
        super().__init__(threads_group="CHAT_POOL", **kwargs)

    def get_input_tokens(self, messages: List[dict]):
        msg_len = count_tokens(json.dumps(messages, ensure_ascii=False))
        msg_len += count_tokens(json.dumps(kwargs.get("tools", []), ensure_ascii=False))
        return msg_len

    def generate(
        self,
        messages: List[dict],
        select: Union[str, Callable]=None,
        **kwargs
    ):
        if callable(select):
            _select = select(self.agents_pool)
        else:
            _select = select

        selected_agent = self.agents_pool.get(_select or list(self.agents_pool.keys())[0], None)
        if not selected_agent:
            raise ValueError(f"Agent not found in pool by {select}")

        yield from selected_agent.generate(messages, **kwargs)




from typing import Union, List, Callable

from .....utils import extract_segments, filter_kwargs, raise_invalid_params
from .....io import EventBlock
from ...prompt_template import PromptTemplate
from ...selector import End
from ..base import BaseAgent
from ..chat import ChatAgent
from .base import FlowAgent

import json

class ReWOO(FlowAgent):
    """
    ReWOO 在 ReAct 基础上, 一次性生成所有需要的计划, 但取消观察环节, 从而大大减少与LLM的对话次数。
    """
    @classmethod
    def allowed_params(cls):
        return {
            "planner": "计划器，用于生成完成任务所有需要的整体计划",
            "solver": "求解器，根据计划和各步骤执行结果汇总后得出最终答案",
            "planner_template": "计划器提示语模板, 默认为 PromptTemplate('FLOW/ReWOO/Planner')",
            "solver_template": "求解器提示语模板, 默认为 PromptTemplate('FLOW/ReWOO/Solver')",
            **FlowAgent.allowed_params(),
        }
        
    def __init__(
        self,
        planner: BaseAgent,
        solver: BaseAgent,
        planner_template: PromptTemplate=None,
        solver_template: PromptTemplate=None,
        **kwargs
    ):
        raise_invalid_params(kwargs, self.allowed_params())

        if not isinstance(planner, ChatAgent):
            raise ValueError("planner 必须是 ChatAgent 的子类")
        if not planner.tools:
            raise ValueError("planner 必须包含可用的工具")
        if solver is planner:
            raise ValueError("planner 和 solver 不能相同")

        planner_template = planner_template or PromptTemplate("FLOW/ReWOO/Planner")
        self.planner_template = planner_template

        solver_template = solver_template or PromptTemplate("FLOW/ReWOO/Solver")
        self.solver_template = solver_template

        self.planner = planner
        self.solver = solver
        self.completed_work = []

        def observe_func(agent: BaseAgent):
            if agent.tools_calling_steps:
                for step in agent.tools_calling_steps:
                    self.completed_work.append(
                        f'Step{step["index"]}: {step["description"]}\n{step["eid"]} = {step["name"]}[{step["arguments"]}]\n{step["eid"]} 执行结果: {step["result"]}\n'
                    )
                self.solver_template.bind_provider({
                    "completed_work": "\n".join(self.completed_work)
                })

                return self.task

        super().__init__(
            planner,
            observe_func,
            solver,
            End(),
            **filter_kwargs(kwargs, self.allowed_params())
        )
    
    def begin_call(self, args):
        self.planner.tools_behavior = "parse-execute"
        self.planner.reset_init_memory(self.planner_template)
        self.planner.clear()

        self.solver.tools_behavior = "nothing"
        self.solver.reset_init_memory(self.solver_template)
        self.solver.clear()

        self.solver_template.bind_provider({
            "completed_work": ""
        })

        return super().begin_call(args)

    def end_call(self):
        self._last_output = self.solver.final_answer



from typing import Any, List
from http import HTTPStatus

import os
from ...types import BaseEmbeddings
from ...utils import raise_invalid_params

class OpenAIEmbeddings(BaseEmbeddings):
    """支持最新的OpenAI文本向量模型"""

    @classmethod
    def allowed_params(cls):
        return {
            "model": "模型名称，默认 'text-embedding-ada-002'",
            "base_url": "BASE_URL",
            "api_key": "API_KEY",
            "dim": "编码时使用的向量维度，默认 1536",
            **BaseEmbeddings.allowed_params()
        }

    def __init__(
        self,
        model: str=None,
        base_url: str=None,
        api_key: str=None,
        dim: int=None,
        extra_args: dict={},
        **kwargs
    ):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        super().__init__(
            model=model or "text-embedding-ada-002",
            base_url=base_url or os.getenv("OPENAI_BASE_URL"),
            api_key=api_key or os.getenv("OPENAI_API_KEY"),
            dim=dim or 1536,
            **kwargs
        )
        try:
            from openai import OpenAI
        except ImportError:
            raise ImportError(
                "Could not import openai package. "
                "Please install it via 'pip install -U openai'"
            )
        
        self.client = OpenAI(base_url=self.base_url, api_key=self.api_key, **extra_args)


    def query(self, text: str,  **kwargs) -> List[float]:
        """
        查询文本向量。
        """
        return self.embed_documents([text], **kwargs)[0]

    def embed_documents(self, texts: List[str], **kwargs) -> List[List[float]]:
        """
        编码文本向量。
        """
        response = self.client.embeddings.create(
            model=self.model,
            input=texts,
            **kwargs
        )
        return [ed.embedding for ed in response.data]




from .prompt import (
    load_resource_template,
    load_prompt_template,
    clone_prompt_template,
    get_template_variables
)

__all__ = [
    "load_resource_template",
    "load_prompt_template",
    "clone_prompt_template",
    "get_template_variables"
]


from typing import List, Union, Optional
from ..base import Runnable
from ....core.document import Document, convert_to_documents_list
from ....utils import hash_text, clean_filename, raise_invalid_params
from ....config import get_env
from ....io import EventBlock

import os
import pickle

class BaseReranker(Runnable):
    """
    重排序器。
    ```
    """
    @classmethod
    def allowed_params(cls):
        return {
            "model": "模型名称",
            "base_url": "API 基础 URL",
            "api_key": "API 密钥",
            "top_k": "返回结果的条数，默认 5",
            **Runnable.allowed_params()
        }

    def __init__(self, model: str=None, base_url: str=None, api_key: str=None, top_k: int=None, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        super().__init__(**kwargs)
        self.model = model
        self.base_url = base_url
        self.api_key = api_key
        self.top_k = top_k or 5
        self.clear_output()

    def clear_output(self):
        self._last_output = []

    def rerank(self, query: str, docs: Union[str, List[str], List[Document]], **kwargs):
        """
        重排序器。
        """
        return docs

    def call(
        self,
        query: str,
        docs: Union[str, List[str], List[Document]],
        **kwargs
    ) -> List[Document]:
        return self.rerank(query, docs)



import json
import os
import asyncio

from typing import Union, List, Optional, Dict, Any
from http import HTTPStatus

from ...utils import raise_invalid_params
from ...io import EventBlock, NewLineBlock
from ...types import ChatAgent
from ..http import confirm_upload_file
from ...io import NewLineBlock

class ChatQwen(ChatAgent):
    @classmethod
    def allowed_params(cls):
        return {
            "model": "模型名称",
            "enable_search": "是否启用搜索",
            "api_key": "API_KEY",
            "base_url": "BASE_URL",
            **ChatAgent.allowed_params()
        }

    def __init__(self, model: str=None, enable_search: bool=False, api_key: str=None, base_url: str=None, extra_args: dict={}, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        try:
            import dashscope
            self.dashscope = dashscope
        except ImportError:
            raise ImportError(
                "Could not import dashscope package. "
                "Please install it via 'pip install -U dashscope'"
            )

        super().__init__(threads_group="CHAT_QWEN", **kwargs)

        self.default_call_args = {
            "model": model or "qwen-plus",
            "enable_search": enable_search
        }
        self.model_args = {
            "api_key": api_key or os.getenv("DASHSCOPE_API_KEY"),
            "base_url": base_url or os.getenv("DASHSCOPE_BASE_URL"),
            **extra_args
        }

    def _prepare_kwargs(self, messages: List[dict], **kwargs) -> dict:
        self.dashscope.api_key = self.model_args["api_key"]

        _kwargs = {
            **self.model_args,
            **self.default_call_args
        }
        _kwargs.update({
            "messages": messages,
            **kwargs,
            **{
                "stream": True,
                "result_format": 'message',
                "incremental_output": True,
            }
        })
        return _kwargs

    def generate(self, messages: List[dict], **kwargs):
        _kwargs = self._prepare_kwargs(messages, **kwargs)

        # 调用生成接口
        responses = self.dashscope.Generation.call(**_kwargs)

        # 流输出
        request_id = None
        usage = {}
        output = []
        for response in responses:
            if response.status_code == HTTPStatus.OK:
                if 'usage' in response:
                    request_id = response.request_id
                    usage = response.usage
                ai_output = response.output.choices[0].message
                output.append(ai_output)
                if 'tool_calls' in ai_output:
                    for func in ai_output.tool_calls:
                        yield EventBlock("tools_call_chunk", json.dumps(func, ensure_ascii=False))
                else:
                    content = ai_output.content
                    yield EventBlock("chunk", content)
            else:
                yield EventBlock("warn", ('Request id: %s, Status code: %s, error code: %s, error message: %s' % (
                    response.request_id, response.status_code,
                    response.code, response.message
                )))

        yield NewLineBlock()
        yield EventBlock(
            "usage",
            json.dumps(usage, ensure_ascii=False),
            calling_info={
                "request_id": request_id,
                "input": _kwargs,
                "output": output,
            }
        )

    async def async_generate(self, messages: List[dict], **kwargs):
        _kwargs = self._prepare_kwargs(messages, **kwargs)

        # 调用生成接口
        responses = await self.dashscope.AioGeneration.call(**_kwargs)

        # 流输出
        usage = {}
        output = []
        request_id = None
        async for response in responses:
            if response.status_code == HTTPStatus.OK:
                if 'usage' in response:
                    request_id = response.request_id
                    usage = response.usage
                ai_output = response.output.choices[0].message
                output.append(ai_output)
                if 'tool_calls' in ai_output:
                    for func in ai_output.tool_calls:
                        yield EventBlock(
                            "tools_call_chunk",
                            json.dumps(func, ensure_ascii=False)
                        )
                else:
                    content = ai_output.content
                    yield EventBlock(
                        "chunk",
                        content
                    )
            else:
                yield EventBlock("warn", ('Request id: %s, Status code: %s, error code: %s, error message: %s' % (
                    response.request_id, response.status_code,
                    response.code, response.message
                )))

        yield NewLineBlock()
        yield EventBlock(
            "usage",
            json.dumps(usage, ensure_ascii=False),
            calling_info={
                "request_id": request_id,
                "input": _kwargs,
                "output": output,
            }
        )

class ChatQwenVL(ChatQwen):
    def __init__(self, model: str=None, **kwargs):
        super().__init__(model=(model or "qwen-vl-plus"), style="qwen_vl", **kwargs)

    def generate(self, messages: List[dict], **kwargs):
        for m in messages:
            if isinstance(m['content'], list):
                for obj in m['content']:
                    if isinstance(obj, dict):
                        if "image" in obj:
                            obj["image"] = confirm_upload_file(self.default_call_args['model'], obj["image"], self.model_args['api_key'])
                        # TODO: 上传视频
                        # 千问模型暂时仅支持上传图片序列的方式实现视频理解，暂不在此支持
                        # elif "video" in obj:
                        #     obj["video"] = confirm_upload_file(self.default_call_args['model'], obj["video"], self.model_args['api_key'])
        _kwargs = self._prepare_kwargs(messages, **kwargs)
        _kwargs["headers"] = {"X-DashScope-OssResourceResolve": "enable"}

        # 调用生成接口
        responses = self.dashscope.MultiModalConversation.call(**_kwargs)

        # 流输出
        usage = {}
        output = []
        request_id = None
        for response in responses:
            if response.status_code == HTTPStatus.OK:
                if 'usage' in response:
                    request_id = response.request_id
                    usage = response.usage
                ai_output = response.output.choices[0].message
                if 'tool_calls' in ai_output:
                    for func in ai_output.tool_calls:
                        output.append({"tools_call_chunk": func})
                        yield EventBlock("tools_call_chunk", json.dumps(func, ensure_ascii=False))
                else:
                    content = ai_output.content
                    output.append({"chunk": content})
                    yield EventBlock("chunk", content)
            else:
                yield EventBlock("warn", ('Request id: %s, Status code: %s, error code: %s, error message: %s' % (
                    response.request_id, response.status_code,
                    response.code, response.message
                )))

        yield NewLineBlock()
        yield EventBlock(
            "usage",
            json.dumps(usage, ensure_ascii=False),
            calling_info={
                "request_id": request_id,
                "output": output,
                "input": _kwargs,
            }
        )

    async def async_generate(self, messages: List[dict], **kwargs):
        for block in await self.run_in_executor(self.generate, messages, **kwargs):
            yield block



from typing import Dict, Any

from ...utils import minify_text, raise_invalid_params, filter_kwargs
from ...hub import load_resource_template, load_prompt_template, get_template_variables
from ...io import EventBlock
from .base import Runnable
from chevron.renderer import render as mustache_render

class PromptTemplate(Runnable):
    """
    提示语模板可以作为消息生成消息列表。
    结合绑定映射，可以动态填充提示语模板。

    dynamic_binding_map 可用于在构造消息时实现动态绑定。
    """
    @classmethod
    def allowed_params(cls):
        return {
            "template_id": "模板 id",
            "text": "模板文本",
            "binding_map": "绑定映射",
            **Runnable.allowed_params()
        }
        
    def __init__(
        self,
        template_id: str=None,
        text: str=None,
        binding_map: Dict=None,
        **kwargs
    ):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        super().__init__(**kwargs)

        if template_id:
            self.text = load_prompt_template(template_id)
        elif text:
            self.text = text
        else:
            raise ValueError('text or template_id cannot be empty')

        if binding_map:
            self.bind_provider(binding_map=binding_map)
        self.template_vars = get_template_variables(self.text)

    def __str__(self):
        return self.format()

    def __repr__(self):
        return f"<PromptTemplate consumer_dict={self.template_vars} text='{minify_text(self.text)}'>"

    def call(self, *args, **kwargs):
        self._last_output = self.format(*args, **kwargs)
        yield EventBlock("info", self._last_output)

    def format(self, binding_map: Dict[str, Any]=None):
        if binding_map:
            self.bind_provider(binding_map=binding_map, dynamic=True)
        _binding = self.consumer_dict
        return mustache_render(template=self.text, data=_binding)




from .base import Runnable
from ...io import EventBlock
from ...utils import minify_text

class Importer(Runnable):
    """
    文档导入。
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def call(self, text: str, **kwargs):
        self._last_output = text
        yield EventBlock("chunk", minify_text(text))



import json
from typing import AsyncIterable, Union
import logging
import asyncio

from .block import EventBlock

def mask_sensitive_info(data):
    """
    遮盖 data 中的敏感信息。
    """
    def mask_value(value):
        if len(value) > 12:
            masked_length = len(value) - 18
            return f"{value[:10]}{'*' * masked_length}{value[-8:]}"
        return value

    if data:
        for key in ["api_key", "base_url"]:
            if key in data and data[key]:
                data[key] = mask_value(data[key])
    return data

def usage(block, verbose: bool=False, **kwargs):
    """
    记录 Runnable 的账单。
    """
    if isinstance(block, EventBlock) and block.block_type == "usage":
        if block.calling_info and "input" in block.calling_info:
            block.calling_info["input"] = mask_sensitive_info(block.calling_info["input"])
        if block.runnable_info:
            block.runnable_info = mask_sensitive_info(block.runnable_info)
        print(block.json)

async def async_usage(block, verbose: bool=False, **kwargs):
    """
    记录 Runnable 的账单。
    """
    if isinstance(block, EventBlock) and block.block_type == "usage":
        if block.calling_info and "input" in block.calling_info:
            block.calling_info["input"] = mask_sensitive_info(block.calling_info["input"])
        if block.runnable_info:
            block.runnable_info = mask_sensitive_info(block.runnable_info)
        print(block.json)
    await asyncio.sleep(0)



from .core.runnable import Runnable, Selector, BaseEmbeddings, VectorDB, PromptTemplate, BaseReranker
from .core.runnable.message import Message, Messages
from .core.runnable.agent import BaseAgent, ChatAgent
from .core.runnable.agent.chat.tools_calling import BaseToolCalling, ToolCall, SubTask, Plans
from .core.runnable.agent.tool_ability import ToolAbility
from .core.document import Document
from .core.dataset import Dataset
from .core.team import Team
from .io import EventBlock, EndBlock

__all__ = [
    "Runnable",
    "Selector",
    "BaseAgent",
    "ChatAgent",
    "BaseEmbeddings",
    "BaseReranker",
    "VectorDB",
    "Message",
    "Messages",
    "PromptTemplate",
    "Document",
    "EventBlock",
    "EndBlock",
    "Dataset",
    "Team",
]



from typing import List, Union
from ...core.runnable.reranker.base import BaseReranker
from ...core.document import Document, convert_to_documents_list
from ...config import get_env
from ...io import EventBlock, NewLineBlock
from ...utils import raise_invalid_params

import os
import json
from http import HTTPStatus


class DashScopeReranker(BaseReranker):
    @classmethod
    def allowed_params(cls):
        return {
            "model": "模型名称",
            "api_key": "API_KEY",
            "top_k": "返回结果的条数，默认 5",
            **BaseReranker.allowed_params()
        }

    def __init__(self, model: str=None, api_key: str=None, top_k: int=None, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        try:
            import dashscope
            self.dashscope = dashscope
        except ImportError:
            raise ImportError(
                "Could not import dashscope package. "
                "Please install it via 'pip install -U dashscope'"
            )

        super().__init__(
            model=model or "gte-rerank",
            api_key=api_key or os.getenv("DASHSCOPE_API_KEY"),
            top_k=top_k if top_k is not None else 5,
            **kwargs
        )

    def rerank(self, query: str, docs: Union[str, List[str], List[Document]], top_k: int=None, **kwargs):
        """
        重排序器。

        rerank 返回的文档长度应当与 docs 一致，因此 top_k 应当取值为 docs 的长度。
        """
        self._last_output = []
        docs = convert_to_documents_list(docs)

        self.dashscope.api_key = self.api_key

        resp = self.dashscope.TextReRank.call(
            model=self.model,
            query=query,
            documents=[doc.text for doc in docs],
            top_n=len(docs),
            return_documents=False
        )

        request_id = None
        usage = {}
        final_docs = []
        if resp.status_code == HTTPStatus.OK:
            request_id = resp.request_id
            if "usage" in resp:
                usage = resp.usage
            if "output" in resp:
                output = resp.output
                if "results" in output:
                    results = output["results"]
                    results.sort(key=lambda x: x["index"], reverse=False)
                    for index, result in enumerate(results):
                        doc = docs[index]
                        doc.meta["rerank_score"] = result["relevance_score"]
                        final_docs.append(doc)
                    final_docs.sort(key=lambda x: x.meta["rerank_score"], reverse=True)
                    self._last_output = final_docs[:(top_k or self.top_k)]

        else:
            yield EventBlock("warn", ('Request id: %s, Status code: %s, error code: %s, error message: %s' % (
                resp.request_id, resp.status_code,
                resp.code, resp.message
            )))

        yield NewLineBlock()
        yield EventBlock(
            "usage",
            json.dumps(usage, ensure_ascii=False),
            calling_info={
                "request_id": request_id,
                "input": {"model": self.model, "query": query, "documents": docs},
                "output": json.dumps(output, ensure_ascii=False),
            }
        )



import os
import shutil
from typing import Any, Set, Union, List

from ....config import get_env
from ...document import Document
from ..vectordb import VectorDB
from ...team import Team
from .retriever import Retriever

class KnowledgeManager:
    @classmethod
    def allowed_params(cls):
        """
        返回当前可用的参数列表。
        """
        return {
            "knowledge": "待检索的资料或向量数据库",
            "team": "所属团队"
        }

    def __init__(
        self,
        knowledge: Union[Set[Any], List[Any]] = None,
        team: Team = None,
    ):
        """
        知识库在内存中以集合的方式保存，确保唯一性。

        默认情况下，会将提供的第一个向量数据库作为默认向量库，默认向量库将自动加载 __ILLUFLY_DOCS__ 和 __ILLUFLY_CHAT_LEARN__ 目录下的文档。
        除非在其他向量库中已经指定了如何加载这两个目录。
        """
        self.knowledge = knowledge

        # 所属团队
        self.team = team

        if isinstance(knowledge, list):
            self.knowledge = set(knowledge)

        if not isinstance(self.knowledge, set):
            self.knowledge = set({self.knowledge}) if self.knowledge else set()

        self.default_docs = set({})
        self.default_vdb = None

        self._recalled_knowledge = []

        self.load_default_knowledge()
        self.load_resource_knowledge()

    @property
    def recalled_knowledge(self):
        """
        返回最近一次调用 get_knowledge 方法时返回的资料列表。
        """
        return self._recalled_knowledge

    def load_default_knowledge(self):
        """
        加载默认知识库。
        """
        self.default_docs = set({
            get_env("ILLUFLY_DOCS"),
            self.team.chat_learn_folder if self.team else get_env("ILLUFLY_CHAT_LEARN")
        })
        for item in self.knowledge:
            if not isinstance(item, (str, Document, VectorDB, Retriever)):
                raise ValueError("Knowledge list items MUST be str, Document or VectorDB")

            if isinstance(item, VectorDB):
                if not self.default_vdb:
                    self.default_vdb = item
                if item in item.sources:
                    # 如果已经在向量库中指定了文档目录，则不再从默认文档目录中加载
                    self.default_docs.remove(item)

        if self.default_vdb:
            for doc_folder in self.default_docs:
                self.default_vdb.load(dir=doc_folder)

    def _get_resource_type(self, ext: str):
        return {
            "jpg": "图片",
            "png": "图片",
            "mp4": "视频",
        }.get(ext, "其他文件")

    def load_resource_knowledge(self, resource_dir: str=None, extensions: List[str]=None):
        """
        自动扫描资源目录中的可观测资源，将其中的图片、视频、音频等可作为资源使用的文件信息提炼为知识，添加到知识库中。
        """
        extensions = extensions or ["jpg", "png", "mp4", "mp3", "wav"]
        resource_docs = resource_dir or get_env("ILLUFLY_RESOURCE")
        if resource_docs:
            if os.path.exists(resource_docs):
                for root, dirs, files in os.walk(resource_docs):
                    # 过滤掉.开头的文件夹
                    dirs[:] = [d for d in dirs if not d.startswith('.')]
                    for file in files:
                        if not file.startswith('.'):  # 过滤掉.开头的文件
                            ext = os.path.splitext(file)[1][1:]
                            if ext in extensions:
                                name = os.path.splitext(file)[0]
                                desc = f"{self._get_resource_type(ext)}资源名: {os.path.join(root, file)}, 资源说明: {name}"
                                self.add_knowledge(desc)

    def add_resource(self, resource_name: str, description: str=None):
        """
        添加资源信息到知识库中。
        """
        ext = os.path.splitext(resource_name)[1][1:]
        resource_type = self._get_resource_type(ext)
        self.add_knowledge(f"资源名: {resource_name}, 资源类型: {resource_type}, 资源说明: {description}")

    def add_knowledge(self, item: Union[str, Document, VectorDB, Retriever]):
        """
        添加知识信息到知识库中。
        """
        if isinstance(item, (str, Document, VectorDB, Retriever)):
            self.knowledge.add(item)
        else:
            raise ValueError("Knowledge MUST be a string, Document or VectorDB")

    def get_knowledge(self, query: str=None, verbose: bool=False):
        """
        根据知识清单召回知识，并返回知识文本列表。
        """
        knowledge = []
        self._recalled_knowledge.clear()
        for kg in self.knowledge:
            if isinstance(kg, Document):
                knowledge.append(kg.text)
                self._recalled_knowledge.append(kg)
            elif isinstance(kg, str):
                knowledge.append(kg)
                self._recalled_knowledge.append(kg)
            elif isinstance(kg, (VectorDB, Retriever)):
                docs = kg(query, verbose=verbose)
                self._recalled_knowledge.extend(docs)
                knowledge.append("\n\n".join([doc.text for doc in docs]))
            else:
                raise ValueError("Knowledge MUST be a string, Document or VectorDB")
        return knowledge

    def clone_chat_learn(self, dest: str, src: str=None):
        """
        克隆 illufly 自身的聊天问答经验。
        """
        if not os.path.exists(dest):
            os.makedirs(dest)

        files_count = 0
        src = src or get_env("ILLUFLY_CHAT_LEARN")
        for item in os.listdir(src):
            if item.startswith('.'):
                continue
            src_path = os.path.join(src, item)
            dst_path = os.path.join(dest, item)
            if os.path.isdir(src_path):
                # 如果是目录，递归拷贝
                self.clone_chat_learn(dst_path, src_path)
            else:
                # 如果是文件，直接拷贝
                shutil.copy2(src_path, dst_path)
                files_count += 1
        return f"从 {src} 拷贝到 {dest} 完成，共克隆了 {files_count} 个文件。"

    def clear_chat_learn(self):
        """
        清空聊天问答经验。
        """
        src = get_env("ILLUFLY_CHAT_LEARN")
        if os.path.exists(src):
            shutil.rmtree(src)
